#
# $Id: mcSSP.pm,v 1.11 2003/04/19 23:52:46 mcooper Exp $
#
# SysInfo Service Protocol (SSP) module
#

package mcSSP;
require 5.6.1;

use mcGlobals;
use mcMisc;
use mcDebug;

use IO::Socket::INET;			# Socket module

#
# SSP Values from mcsysinfod/include/ssp.h
#
$SSP_TCP_PORT  		= 11967;	# Standard SSP port
# Commands
$SSP_CMD_AUTH		= 'AUTH';	# Send auth info
$SSP_CMD_BYE		= 'BYE';	# Shutdown connection
$SSP_CMD_CLASSLIST	= 'CLASSLIST';	# List available classes
$SSP_CMD_LICENSE	= 'LICENSE';	# Show license info
$SSP_CMD_SEND		= 'SEND';	# Send data
$SSP_CMD_SETOPTION	= 'SETOPTION';	# Set options
$SSP_CMD_VERSION	= 'VERSION';	# Send/receive version info
$SSP_CMD_MEMBERS	= 'MEMBERS';	# Send members info
$SSP_CMD_TYPES		= 'TYPES';	# Send types info
$SSP_CMD_NOOP		= 'NOOP';	# Do nothing
# Auth types
$SSP_AUTH_PASSWORD	= 'PASSWORD';	# Username + password
# Response Codes
$SSP_RC_OK		= 'OK';		# Success
$SSP_RC_ERR		= 'ERR';	# Error
$SSP_RC_AUTHREQ		= 'AUTHREQ';	# Authentication required
$SSP_RC_START		= 'START';	# Start data transmit
$SSP_RC_BYE		= 'BYE';	# End session
# Response continuation chars
$SSP_RCC_SINGLE		= ' ';		# Single line response
$SSP_RCC_MULTI		= '-';		# Multi line response

# Default timeout for reading 1 line
$TimeoutDefault		= 5 * 60;
# Timeout upon initial connection startup
$TimeoutStartup		= 15;

#
# Table of all $SSP_RC_* codes
#
my $SSPcodeTable =
{
 $SSP_RC_OK 		=> { desc=>'Ready' },
 $SSP_RC_ERR		=> { desc=>'Error' },
 $SSP_RC_START		=> { desc=>'Start of data stream' },
 $SSP_RC_AUTHREQ	=> { desc=>'Authentication required' },
 $SSP_RC_BYE		=> { desc=>'Ending session' },
};

#
# Object Constructor
#
# Make a connect to $Host and $Port.
#
# $Args->{Username} and $Args->{Password} should be present if
# server asks for authentication.
#
# Return $self on success.
# Return undef on error.
#
sub New {
    my $proto = shift;
    my $Args = shift;
    my $Host = $Args->{Host};
    my $Port = $Args->{Port} || $SSP_TCP_PORT;
    my $class = ref($proto) || $proto;
    my $self = {};

    if (!$Args || !$Args->{Host}) {
	Error("mcSSP->New: Missing parameters.");
	return undef;
    }

    $self->{"Retrieve"} = undef;
    $self->{"Host"} = undef;
    $self->{"Socket"} = undef;
    $self->{"Timeout"} = undef;		# Server timeout (in seconds)
    $self->{"FH"} = undef;
    $self->{"GetLine"} = undef;
    $self->{"Info"} = [];		# Array of Info from server
    $self->{"Command"} = undef;		# Command used for this entry
    $self->{"Startup"} = undef;
    $self->{"Status"} = undef;		# Current status

    bless ($self, $class);

    #
    # Make the connection
    #
    my $socket = IO::Socket::INET->new(PeerHost=>$Host,
				       PeerPort=>$Port,
				       Proto=>'tcp',
				       Type=>SOCK_STREAM);
    if (!$socket) {
	Error("Connect to \"%s\" on port %d failed: %s",
	      $Host, $Port, $!);
	if (Eq($!, 'Connection refused')) {
	    Error("The SysInfo server does not appear to be running on %s.",
		  $Host);
	}
	return undef;
    }

    # So far so good
    $self->Socket($socket);
    $self->Host($Host);

    # Startup the connection to the server
    if ($self->Startup($Args) < 0) {
	return undef;
    }

    # Success
    return $self;
}

sub Host {
    my $self = shift;

    if (!@_) {
	return $self->{"Host"};
    }

    $self->{"Host"} = shift;

    return;
}

sub Timeout {
    my $self = shift;

    if (!@_) {
	return $self->{"Timeout"};
    }

    $self->{"Timeout"} = shift;

    return;
}

sub FH {
    my $self = shift;

    if (!@_) {
	return $self->{"FH"};
    }

    $self->{"FH"} = shift;

    return;
}

sub Socket {
    my $self = shift;

    if (!@_) {
	return $self->{"Socket"};
    }

    $self->{"Socket"} = shift;

    return;
}

sub Command {
    my $self = shift;

    if (!@_) {
	return $self->{"Command"};
    }

    $self->{"Command"} = shift;

    return;
}

sub Info {
    my $self = shift;

    if (!@_) {
	return @{ $self->{"Info"} };
    }

    @{ $self->{"Info"} } = shift;

    return;
}

#
# Current status.
# Should only handle $SSP_RC_* values.
#
sub Status {
    my $self = shift;

    if (!@_) {
	return $self->{"Status"};
    }

    $self->{"Status"} = shift;

    return;
}

#
# Return a line of info read from server
#
sub ReadLine {
    my $self = shift;
    my $Args = shift;
    my $Timeout = $Args->{Timeout};
    my $Socket = $self->Socket;
    my $Line;

    if (!$Socket) {
	Error("mcSSP->ReadLine: We are not connected to server %s.",
	      $self->Host);
	$self->Down;
	return;
    }

    #
    # Put the read inside an eval with the timer to allow us to receive
    # the SIGALRM even during a system call.  Per the perlfunc(1) man page.
    #
    my $timeout = $Timeout || $self->Timeout || $TimeoutDefault;
    eval {
	local $SIG{ALRM} = sub { die "alarm\n" }; # The \n is required
	alarm $timeout;
	$Line = <$Socket>;
	alarm 0;
    };
    if ($@) {
	if ($@ eq "alarm\n") {
	    Error("Read from server %s timed out after %d seconds.",
		  $self->Host, $timeout);
	    # Mark connection as down
	    $self->Down;
	    return;
	}
    }

    if ($Line =~ /\n/) {
	chop($Line);
    }

    Debug("< %s", $Line);

    return $Line;
}

#
# Parse an SSP response line and return the resulting hash.
#
sub ResultParse {
    my $self = shift;
    my $Line = shift;
    my $Result;

    if (!$Line) {
	return;
    }

    if ($Line !~ /^\*/) {
	Error("mcSSP->ResultParse: Line does not appear to be an SSP RESPONSE: %s",
	      $Line);
	return;
    }

    my @bytes = split(//, $Line);	# Array of bytes
    my $rcc = $bytes[1];		# Response Continuation Code

    # Get the response code
    my $Code = $Line;
    $Code =~ s/^..//;
    $Code =~ s/[ ].*//g;

    my $Info = $Line;
    $Info =~ s/.*$Code[ ]//;

    Debug("mcSSP->ResultParse: Got=<%s>\n\tCode=<%s> RCC=<%s>\n\tOther=<%s>\n",
	  $Line, $Code, $rcc, $Info, $Line);

    $Result->{Code} = $Code;
    $Result->{RCC} = $rcc;
    $Result->{Info} = $Info;

    return $Result;
}

#
# Return the results of a command from the server
# Data is read from $self->Socket.
#
sub Result {
    my $self = shift;
    my $Args = shift;
    my $Timeout = $Args->{Timeout};
    my $Result;

    while (my $Line = $self->ReadLine({Timeout=>$Timeout})) {
	# Check to see if line starts with "*"
	my $LineRes = $self->ResultParse($Line);
	if (!$LineRes) {
	    last;
	}

	# Add to primary (initial) code response if this is our first
	if (!$Result->{Code}) {
	    $Result->{Code} = uc($LineRes->{Code});
	}
	# Add to list of all codes for this result
	push(@{ $Result->{Codes} }, $LineRes->{Code});
	#
	# Add other info from server to array.
	# We add regardless of whether info was present to match numeric
	# order of {Codes} array.
	#
	push(@{ $Result->{Info} }, $LineRes->{Info});

	if ($LineRes->{RCC} ne $SSP_RCC_MULTI) {
	    last;	    # Done with results
	}
    }

    $self->Info(@{$Result->{Info}});

    if (!$Result || !$Result->{Code}) {
	Error("mcSSP->Result: Failed to read an SSP response code from %s",
	      $self->Host);
	return;
    }

    return $Result;
}

#
# Does $Result contain AUTHREQ?
#
sub NeedAuth {
    my $Result = shift;

    foreach my $code (@{ $Result->{Codes} }) {
	if (Eq($code, $SSP_RC_AUTHREQ)) {
	    return 1;
	}
    }

    return 0;
}

#
# Send auth info to the server
#
sub Auth {
    my $self = shift;
    my $Args = shift;

    if (!$Args->{Username} || !$Args->{Password}) {
	Error("mcSSP->Auth: Server \"%s\" requires a username and password.",
	      $self->Host);
	return -1;
    }
    # Send the AUTH command
    if ($self->SendCommand({Cmd=>$SSP_CMD_AUTH,
			    Arg=>join(' ',
				      $SSP_AUTH_PASSWORD,
				      $Args->{Username},
				      $Args->{Password})
			   }) < 0) {
	Error("mcSSP->Auth: Send %s command failed.", $SSP_CMD_AUTH);
	return -1;
    }
    # Check the results of the AUTH
    my $authres = $self->Result;
    if (!Eq($authres->{Code}, $SSP_RC_OK)) {
	Error("The authentication with username \"%s\" to server \"%s\" failed.",
	      $Args->{Username},
	      $self->Host);
	Error("The server said: %s", $authres->{Info}[0]);
	return -1;
    }

    Debug("Authentication with \"%s\" as \"%s\" succeeded.",
	  $self->Host, $Args->{Username});

    return 0;
}

#
# Talk to the server during initial startup of the connection
#
sub Startup {
    my $self = shift;
    my $Args = shift;

    my $result = $self->Result({Timeout=>$TimeoutStartup});;
    if (!$result) {
	return;
    }

    #
    # See if the server told us that we need to authenticate
    #
    if (NeedAuth($result)) {
	Debug("mcSSP->Startup: Server requires AUTH.");
	if ($self->Auth($Args) < 0) {
	    return -1;
	}
    }

    if (Eq($result->{Code}, $SSP_RC_OK)) {
	$self->Status($SSP_RC_OK);
	Debug("mcSSP->Startup: Connection established with SSP server.");
	return 0;
    } else {
	return -1;
    }
}

#
# Send a command to the server
# Returns 0 on success.
# Returns -1 on error.
#
sub SendCommand {
    my $self = shift;
    my $Args = shift;
    my $CmdName = $Args->{Cmd};
    my $Arg = $Args->{Arg};
    my $Socket = $self->Socket;

    if (!$Args || !$CmdName) {
	Error("mcSSP->SendCommand: Missing parameters.");
	return -1;
    }

    if (!$Socket) {
	Error("mcSSP->SendCommand: Socket is not connected.");
	$self->Down;
	return -1;
    }

    my $Cmd = $CmdName;
    if ($Arg) {
	$Cmd = $Cmd . ' ' . $Arg;
    }

    Debug("> %s", $Cmd);

    my $r = syswrite $Socket, $Cmd . "\n";
    if ($r != length($Cmd) + 1) {
	Error("mcSSP->SendCommand: Send command \"%s\" failed: %s",
	      $Cmd, $!);
	return -1;
    }

    return 0;
}

#
# Prepare to retrieve data from server.
# Returns 0 on success -> next call should be to $self->Data
# Returns -1 on error.
#
sub Retrieve {
    my $self = shift;
    my $Args = shift;
    my @ClassList = @{ $Args->{ClassList} };
    my $Info = $Args->{Info};
    my $Options = $Args->{Options};

    if (!$Args || $#ClassList < 0) {
	Error("mcSSP->Retrieve: Missing parameters.");
	return -1;
    }

    # Verify connection is good to go
    if ($self->Status ne $SSP_RC_OK) {
	Error("mcSSP->Retrieve: Connection status to \"%s\" is not \"%s\" (currently \"%s\").",
	      $self->Host, $SSP_RC_OK, $self->Status);
	return -1;
    }

    #
    # Set options
    #
    if ($Options) {
	#
	# @opts should be of form $opt=$val
	#
	my @opts;
	foreach my $o (keys %$Options) {
	    if ($Options->{$o}) {
		# Key=value
		push(@opts, $o . '=' . $Options->{$o});
	    } else {
		# Boolean
		push(@opts, $o);
	    }
	}
	if ($self->SendCommand({Cmd=>$SSP_CMD_SETOPTION,
				Arg=>join(' ', @opts)}) < 0) {
	    Error("mcSSP->Retrieve: Send command \"%s\" to \"%s\" failed.",
		  $SSP_CMD_SETOPTION, $self->Host);
	    return -1;
	}

	# Read the result
	my $result = $self->Result;
	if (!Eq($result->{Code}, $SSP_RC_OK)) {
	    Error("mcSSP->Retrieve: Unexpected error setting options from \"%s\".",
		  $self->Host);
	    return -1;
	}
    }

    my $cmd;
    if ($Info) {
	if (Eq($Info, $RI_MCL)) {
	    $cmd = $SSP_CMD_LICENSE;
	} elsif (Eq($Info, $RI_VERSION)) {
	    $cmd = $SSP_CMD_VERSION;
	} elsif (Eq($Info, $RI_MEMBERS)) {
	    $cmd = $SSP_CMD_MEMBERS;
	} elsif (Eq($Info, $RI_TYPES)) {
	    $cmd = $SSP_CMD_TYPES;
	} else {
	    Error("mcSSP->Retrieve: Info command <%s> is unknown.", $Info);
	    return -1;
	}
	#
	# Send the Info specific request
	#
	if ($self->SendCommand({Cmd=>$cmd}) < 0) {
	    Error("mcSSP->Retrieve: Send command \"%s\" to \"%s\" failed.",
		  $cmd, $self->Host);
	    return -1;
	}
    } else {
	#
	# Send the class request
	#
	$cmd = $SSP_CMD_SEND;
	if ($self->SendCommand({Cmd=>$cmd,
				Arg=>join(',', @ClassList)}) < 0) {
	    Error("mcSSP->Retrieve: Send command to \"%s\" failed.",
		  $self->Host);
	    return -1;
	}
    }

    # Save command
    $self->Command($cmd);

    # Read the result
    my $result = $self->Result;
	
    if (Eq($result->{Code}, $SSP_RC_START)) {
	Debug("mcSSP->Retrieve: Ready for data read via mcSSP->GetLine");
	# Success; we're ready for $self->Data to be called
	return 0;
    } elsif (Eq($cmd, $SSP_CMD_VERSION)) {
	# Success; we're ready for $self->Data to be called
	return 0;
    } else {
	Error("mcSSP->Retrieve: Unexpected error asking for data from \"%s\".",
	      $self->Host);
	return -1;
    }
}

#
# Shutdown the connection
#
sub Down {
    my $self = shift;

    if ($self->Socket) {
	close($self->Socket);
    }
    $self->{"Socket"} = undef;
    $self->{"Status"} = undef;
}

#
# Destroy the connection
#
sub Destroy {
    my $self = shift;

    $self->SendCommand({Cmd=>$SSP_CMD_BYE});
    $self->Down;
}

#
# Read one line of data and return it.
# Returns undef when done or on error.
#
sub GetLine {
    my $self = shift;
    my $Socket = $self->Socket;

    if (Eq($self->Command, $SSP_CMD_VERSION)) {
	# What we want is in Info[0]
	my $VerStr = ($self->Info)[0];
	if (!$VerStr) {
	    # We're done
	    return;
	}
	# Set to empty so when we're called the 2nd time there's nothing
	$self->{Info} = [];
	# $VerStr should look like: SSP=1 SERVER=5.0.0-alpha1 SIE=5.0-alpha1
	my @Parts = split(/[ ]+/, $VerStr);
	# Make $VerInfo into hash
	my $VerInfo;
	foreach my $s (@Parts) {
	    my @p = split('=', $s);
	    if ($#p == 1) {
		$VerInfo->{uc($p[0])} = $p[1];
	    }
	}

	return $VerInfo;
    }

    if (!$Socket) {
	Error("mcSSP->GetLine: We do not appear to be connected.");
	return;
    }

    my $Line = <$Socket>;
    if (!$Line) {
	Debug("mcSSP->GetLine: Server returned nothing unexpectedly.");
	$self->Down;
    }

    chop($Line);

    #
    # See if this is an SSP response code.
    # We're looking to see if the server is done sending data.
    #
    if ($Line =~ /^\*/) {
	my @bytes = split(//, $Line);
	my $rcc = $bytes[1];
	if ($rcc eq $SSP_RCC_SINGLE || $rcc eq $SSP_RCC_MULTI) {
	    # Yup, it's an SSP response
	    my $Result = $self->ResultParse($Line);
	    my $code = uc($Result->{Code});
	    if (Eq($code, $SSP_RC_OK)) {
		# No more data
		return;
	    } elsif ($SSPcodeTable->{$code}) {
		Error("Unexpected server response: %s %s",
		      $code, $Result->{Info}[0]);
		return;
	    }
	    # Must not be a valid command so ignore it
	    Debug("mcSSP->GetLine: Ignoring line which looks like SSP: <%s>",
		  $Line);
	}
    }

    return $Line;
}

1;	# required
