#
# $Id: mcStream.pm,v 1.4 2003/04/21 21:03:39 mcooper Exp $
#
# Stream layer which provides mcSysInfo with data.
#
# This object acts as abstraction layer between mcSysInfo and the
# stream drivers which are the actual data providers.  The Open
# function chooses which driver to call and if to use the Cache driver.
# We also update the cache on Read.
#

package mcStream;
require 5.003;

use mcGlobals;
use mcMisc;
use mcDebug;
use mcStreamSie;
use mcStreamSsp;
use mcStreamFile;
use mcStreamCache;

#
# Object Constructor
#
sub New {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = {};
    my $Err;

    $self->{"Open"} = undef;
    $self->{"Close"} = undef;
    $self->{"ReadLine"} = undef;
    $self->{"Driver"} = undef;
    $self->{"Cache"} = undef;

    bless ($self, $class);

    return $self;
}

sub Driver {
    my $self = shift;

    if (!@_) {
	return $self->{"Driver"};
    }

    $self->{"Driver"} = shift;

    return;
}
sub Cache {
    my $self = shift;

    if (!@_) {
	return $self->{"Cache"};
    }

    $self->{"Cache"} = shift;

    return;
}

#
# Open a stream.
# Returns mcStatus
#
sub Open {
    my $self = shift;
    my $SysInfo = shift;
    my $Args = shift;
    my @Classes;
    my $Driver;

    #
    # Check to see if we should use the SysInfo object passed in $Args
    # instead of the object passed as arg[1].
    #
    if (exists($Args->{SysInfo}) && (my $si = $Args->{SysInfo})) {
	if (!$SysInfo || $si->InputFile || $si->Host) {
	    $SysInfo = $si;
	}
    }

    my $Status = mcStatus->New;

    if ($SysInfo->InputFile) {
	#
	# Retrieve from a file
	#
	$Driver = mcStreamFile->New;
    } else {
	#
	# These stream drivers work with the Cache driver
	#
	my $Cache = mcStreamCache->New;
	$Status = $Cache->Open($SysInfo, $Args);
	if ($Status->Code == $Status->SUCCESS) {
	    # Use the cache as the driver
	    $Driver = $Cache;
	} else {
	    #
	    # The cache is not available or is out of date so go direct
	    #
	    if ($SysInfo->Host) {
		#
		# Retrieve from remote host via SSP server
		#
		$Driver = mcStreamSsp->New;
	    } else {
		#
		# Retrieve directly from local SysInfo Engine.
		#
		$Driver = mcStreamSie->New;
	    }
	    #
	    # Save the Cache driver so we can update the cache as we go
	    #
	    if ($Driver) {
		$self->Cache($Cache);
	    }
	}
    }

    if (!$Driver) {
	Error("Stream->Open: Cannot determine which Stream driver to use.");
	$Status->Code($Status->FAILURE);
	$Status->Value($Status->HARD);	# Hard error
	return $Status;
    }

    $self->Driver($Driver);

    return $Driver->Open($SysInfo, $Args);
}

#
# Close a stream
#
sub Close {
    my $self = shift;

    if (exists($self->{Driver}) && exists($self->Driver->{Close})) {
	$self->Driver->Close;
    }

    if ($self->Cache) {
	$self->Cache->Close;
    }
}

#
# Read a line from $self
#
sub ReadLine {
    my $self = shift;
    my $Line;

    # Read a line of data from the driver
    $Line = $self->Driver->ReadLine;

    # Save this line to the cache if we're not reading from it already
    if ($self->Cache && exists($self->Cache->{FileCache})) {
	$self->Cache->{FileCache}->WriteData($Line . "\n");
    }

    return $Line;
}
