#
# $Id: mcSysInfo.pm,v 1.52 2003/05/15 01:13:42 mcooper Exp $
#
# Main mcSysInfo perl object
#

package mcSysInfo;
require 5.003;

use mcGlobals;
use mcMisc;
use mcDebug;
use mcGeneral;
use mcHardware;
use mcPartition;
use mcPrinter;
use mcSoftware;
use mcLicense;
use mcSysConf;
use mcKernel;
use mcNetIf;
use mcNetwork;
use mcMagniCompLic;
use mcMember;
use mcType;
use mcSSP;
use mcStatus;
use mcStream;

my $SaveSTDERR;

#
# Object Constructor
#
sub New {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = {};
    my $Err;

    $self->{"Program"} = undef;
    $self->{"ProgramArgs"} = [];
    $self->{"InputFile"} = undef;
    $self->{"Host"} = undef;
    $self->{"Port"} = undef;
    $self->{"Username"} = undef;
    $self->{"Password"} = undef;
    $self->{"Class"} = [];
    $self->{"General"} = mcGeneral->New();
    $self->{"Hardware"} = mcHardware->New();
    $self->{"Software"} = mcSoftware->New();
    $self->{"License"} = mcLicense->New();
    $self->{"Partition"} = mcPartition->New();
    $self->{"Printer"} = mcPrinter->New();
    $self->{"NetIf"} = mcNetIf->New();
    $self->{"Network"} = mcNetwork->New();
    $self->{"SysConf"} = mcSysConf->New();
    $self->{"Kernel"} = mcKernel->New();
    $self->{"MagniCompLic"} = mcMagniCompLic->New();
    $self->{"Types"} = undef;
    $self->{"Version"} = undef;

    $self->{"SSP"} = undef;		# Internal use

    $self->{"MembersDecode"} = undef;
    $self->{"TypesDecode"} = undef;

    bless ($self, $class);

    return $self;
}

#
# Destroy ourself
#
sub Destroy {
    my $self = shift;

    if ($self->SSP && exists($self->SSP->{Destroy})) {
	# Time to shutdown the connection
	$self->SSP->Destroy;
    }
}

#
# Name/path of sysinfo program
#
sub Program {
    my $self = shift;

    if (!@_) {
	return $self->{"Program"};
    }

    $self->{"Program"} = shift;

    return;
}

#
# Input File
#
sub InputFile {
    my $self = shift;

    if (!@_) {
	return $self->{"InputFile"};
    }

    $self->{"InputFile"} = shift;

    return;
}

#
# Host to query
#
sub Host {
    my $self = shift;

    if (!@_) {
	return $self->{"Host"};
    }

    $self->{"Host"} = shift;

    return;
}

#
# Port to query
#
sub Port {
    my $self = shift;

    if (!@_) {
	return $self->{"Port"};
    }

    $self->{"Port"} = shift;

    return;
}

#
# Username on Host
#
sub Username {
    my $self = shift;

    if (!@_) {
	return $self->{"Username"};
    }

    $self->{"Username"} = shift;

    return;
}

#
# Password on Host
#
sub Password {
    my $self = shift;

    if (!@_) {
	return $self->{"Password"};
    }

    $self->{"Password"} = shift;

    return;
}

#
# XXX The SSP member is used to allow us to cache the SSP connection
# for the lifetime of this mcSysInfo object.  This avoids multiple
# new connections for each type of requested data.
#
sub SSP {
    my $self = shift;

    if (!@_) {
	return $self->{"SSP"};
    }

    $self->{"SSP"} = shift;

    return;
}

#
# Class info list
#
sub Class {
    my $self = shift;

    if (!@_) {
	my @a = @{ $self->{"Class"} };
	return ($#a < 0) ? @AllClassNames : @a;
    }

    @{ $self->{"Class"} } = @_;

    return;
}

#
# Version of SysInfo
# Returns hash of version info:
#	PRODUCT		- Overall product version
#	SIE		- SysInfo Engine (sie) version
#	SERVER		- Server (mcsysinfod) version
#	SSP		- SSP protocol version
#
sub Version {
    my $self = shift;

    if (!@_) {
	if (!$self->{"Version"}) {
	    $self->Retrieve({Info=>$RI_VERSION});
	}

	#
	# If Version is a scalar (because Retrieve set it)
	# then reset as a hash.
	#
	if (ref($self->{"Version"}) !~ /HASH/) {
	    my $v = $self->{"Version"};
	    # Insure we have no spaces
	    $v =~ s/[ ]/\-/g;
	    my $h;
	    $h->{SIE} = $v;
	    $self->{"Version"} = $h;
	}

	# Insure we have a PRODUCT hash member
	if ($self->{"Version"} && !exists($self->{"Version"}->{PRODUCT}) &&
	    exists($self->{"Version"}->{SIE})) {
	    $self->{"Version"}->{PRODUCT} = $self->{"Version"}->{SIE};
	}

	return $self->{"Version"};
    }

    $self->{"Version"} = shift;

    return;
}

#
# Retrieve Member info
# Returns $self on success.
# Returns undef on failure;
#
sub RetrieveMembers {
    my $self = shift;
    my $Args = shift;

    $Args->{Info} = $RI_MEMBERS;
    $r = $self->Retrieve($Args);
    delete($Args->{Info});

    if (!$r) {
	# Failure
	return;
    }

    # Success
    return $self;
}

#
# Types member is a hash of mcType objects
#
sub Types {
    my $self = shift;
    my $Args = shift;

    if (!@_) {
	if (!$self->{"Types"}) {
	    # Retrieve the data
	    $Args->{Info} = $RI_TYPES;
	    $r = $self->Retrieve($Args);
	    delete($Args->{Info});

	    if (!$r) {
		# Failure
		return;
	    }
	}
	return $self->{"Types"};
    }

    $self->{"Types"} = shift;

    return;
}

#
# Fetch the object named $Name from $self
#
sub GetObjClass {
    my $self = shift;
    my $Name = shift;

    foreach my $n (keys %$self) {
	if (lc($n) eq lc($Name)) {
	    # Success
	    return $self->{$n};
	}
    }

    # Failure
    return;
}

#
# Decode a 'members' entry from @Fields
#
sub MembersDecode {
    my $self = shift;
    my @Fields = @_;

    my $Member = mcMember->New;
    if (!$Member->Decode(@Fields)) {
	# Failure
	return;
    }

    # Use $Member->ClassName to set this entry in the right object class
    my $Obj = $self->GetObjClass($Member->ClassName);
    if (!exists($Obj->{Members})) {
	# Failure
	return;
    }

    # Grab current Members
    my $m = $Obj->{Members};
    # Choose the key to use for saving this member
    my $key = $Member->Name || $Member->Key;
    # Add new $Member using obj member name as key
    $m->{$key} = $Member;
    # Save back
    $Obj->Members({}, $m);

    # Success
    return $self;
}

#
# Decode a 'types' entry from @Fields
#
sub TypesDecode {
    my $self = shift;
    my @Fields = @_;

    my $Type = mcType->New;
    if (!$Type->Decode(@Fields)) {
	# Failure
	return;
    }

    # Grab current Types
    my $m = $self->Types;
    # Choose the key to use for saving this type
    my $key = $Type->Name;
    # Add new $Type object
    $m->{$key} = $Type;
    # Save back
    $self->Types($m);

    # Success
    return $self;
}

#
# Program Arguments
#
sub ProgramArgs {
    my $self = shift;

    if (!@_) {
	return @{ $self->{"ProgramArgs"} };
    }

    @{ $self->{"ProgramArgs"} } = @_;

    return;
}

sub General {
    my $self = shift;

    if (!@_) {
	return $self->{"General"};
    }

    $self->{"General"} = shift;

    return;
}

sub SysConf {
    my $self = shift;

    if (!@_) {
	return $self->{"SysConf"};
    }

    $self->{"SysConf"} = shift;

    return;
}

sub Kernel {
    my $self = shift;

    if (!@_) {
	return $self->{"Kernel"};
    }

    $self->{"Kernel"} = shift;

    return;
}

sub Hardware {
    my $self = shift;

    if (!@_) {
	return $self->{"Hardware"};
    }

    $self->{"Hardware"} = shift;

    return;
}

sub Software {
    my $self = shift;

    if (!@_) {
	return $self->{"Software"};
    }

    $self->{"Software"} = shift;

    return;
}

sub License {
    my $self = shift;

    if (!@_) {
	return $self->{"License"};
    }

    $self->{"License"} = shift;

    return;
}

sub Partition {
    my $self = shift;

    if (!@_) {
	return $self->{"Partition"};
    }

    $self->{"Partition"} = shift;

    return;
}

sub Printer {
    my $self = shift;

    if (!@_) {
	return $self->{"Printer"};
    }

    $self->{"Printer"} = shift;

    return;
}

sub NetIf {
    my $self = shift;

    if (!@_) {
	return $self->{"NetIf"};
    }

    $self->{"NetIf"} = shift;

    return;
}

sub Network {
    my $self = shift;

    if (!@_) {
	return $self->{"Network"};
    }

    $self->{"Network"} = shift;

    return;
}

sub MagniCompLic {
    my $self = shift;

    if (!@_) {
	if (!$self->{"MagniCompLic"}) {
	    $self->Retrieve({Info=>$RI_MCL});
	}
	return $self->{"MagniCompLic"};
    }

    $self->{"MagniCompLic"} = shift;

    return;
}

#
# Find SysInfo
#
sub FindSysInfo {
    my $self = shift;

    my $Program = $self->Program;
    if ($Program =~ /\//) {
	Debug("FindSysInfo: Using SysInfo->Program = <%s>", $Program);
	return $Program;
    }

    if (!length($Program)) {
	$Program = "sysinfo";
    }

    foreach my $Dir (split(':', $ENV{'PATH'})) {
	my $p = $Dir . "/" . $Program;
	Debug("FindSysInfo: Checking <%s>", $p);
	if (-x $p) {
	    return $p;
	}
    }

    Error("Cannot locate %s program in \$PATH.", $Program);
    return;
}

#
# Prepare for a new Report run
#
sub ReportSet {
    my $self = shift;
    my $Opts = shift;

    $self->ReportEnd;

    # Setup read function
    my $Stream = mcStream->New;
    my $Status = $Stream->Open($self, $Opts);
    if ($Status->Code == $Status->SUCCESS) {
	$self->{Stream} = $Stream;
    }
}

#
# End Report
#
sub ReportEnd {
    my $self = shift;

    if ($self->{Stream}) {
	$self->{Stream}->Close;
	delete($self->{Stream});
    }
}

#
# Return 1 Report Line
# Returns undef on EOF.
#
sub ReportLine {
    my $self = shift;

    my $Stream = $self->{Stream};
    if (!$Stream) {
	return;
    }

    my $Line = $Stream->ReadLine;

    return $Line;
}

#
# Retrieve data, either $Info or $Class
# Returns 0 on success.
# Returns -1 on soft error.
# Returns -2 on hard (do not retry) error.
#
sub RetrieveData {
    my $self = shift;
    my $Class = shift;
    my $Args = shift;
    my $Info = $Args->{Info};

    Debug("SysInfo->RetrieveData:\t Class=<%s> Info=<%s>", $Class, $Info);

    if ($Class) {
	$Args->{Class} = $Class;	# Set temporary override
    }
    my $Stream = mcStream->New;
    my $Status = $Stream->Open($self, $Args);
    $Args->{Class} = undef;	# Unset temporary override
    if (!$Status) {
	Debug("SysInfo->RetrieveData: \$Stream->Open (obj=%s) returned no Status.",
	     $Stream);
	return -1;
    }
    if ($Status->Code == $Status->FAILURE) {
	if ($Status->Value == $Status->SOFT) {
	    return -1;	# Soft error
	} elsif ($Status->Value == $Status->HARD) {
	    return -2;	# Hard error
	} else {
	    return -1;
	}
    }

    while (my $Line = $Stream->ReadLine) {
	# Select seperator
	my $Sep = '\|';
	if ($Line =~ / \| /) {
	    $Sep = ' \| ';
	}
	if ($Line !~ /$Sep/ && !$Info) {
	    # Input lines must have at least 1 $Sep char
	    next;
	}

	my @Fields = split($Sep, $Line);
	my $class = $Fields[0];

	# Special mapping
	if (Eq($class, 'device')) {
	    $class = 'hardware';
	}

	if ($Info) {
	    #
	    # Decode special Info type
	    #
	    my $decode = $Info . 'Decode';
	    if (exists($self->{$Info}) && exists($self->{$Info}->{Decode})) {
		# Have the object's Decode function handle
		$self->{$Info}->Decode(@Fields);
	    } elsif (exists($self->{$decode})) {
		$self->$decode(@Fields);
	    } elsif (exists($self->{$Info})) {
		# Save as scalar?
		$self->{$Info} = $Line;
	    }
	    next;
	}

	# Skip it if this class is not enabled
	if (!Eq($Class, $class)) {
	    next;
	}

	if (Eq($class, 'general')) {
	    $self->General->Decode(@Fields);
	} elsif (Eq($class, 'hardware')) {
	    $self->Hardware->Decode(@Fields);
	} elsif (Eq($class, 'software')) {
	    $self->Software->Decode(@Fields);
	} elsif (Eq($class, 'license')) {
	    $self->License->Decode(@Fields);
	} elsif (Eq($class, 'partition')) {
	    $self->Partition->Decode(@Fields);
	} elsif (Eq($class, 'printer')) {
	    $self->Printer->Decode(@Fields);
	} elsif (Eq($class, 'netif')) {
	    $self->NetIf->Decode(@Fields);
	} elsif (Eq($class, 'network')) {
	    $self->Network->Decode(@Fields);
	} elsif (Eq($class, 'sysconf')) {
	    $self->SysConf->Decode(@Fields);
	} elsif (Eq($class, 'kernel')) {
	    $self->Kernel->Decode(@Fields);
	} else {
	    Debug("SysInfo->Retrieve: Class <%s> is undefined.", $class);
	}
    }

    $Stream->Close;

    return 0;
}

#
# Retrieve the requested data.
# Valid values for Info are:
#	$RI_MCL		- Retrieve mcMagniCompLic object
#	$RI_MEMBERS	- Retrieve mcMember info for each class
#	$RI_VERSION	- Get version of SysInfo
#	$RI_TYPES	- Retrieve mcSysInfo->Types
#
sub Retrieve {
    my $self = shift;
    my $Args = shift;
    my @ClassList = @{ $Args->{ClassList} };
    my $Info = $Args->{Info};

    Debug("SysInfo->Retrieve: Retrieving SysInfo data...");

    # Update Class member if not set
    if (!$self->Class && $Args->{Classes}) {
	$self->Class($Args->{Classes});
    }

    my @Classes = $self->Class;
    # Check to see if caller specified classes via ClassList
    if ($#ClassList >= 0) {
	@Classes = @ClassList;
    }

    Debug("SysInfo->Retrieve:\t Classes=<%s> Info=<%s>",
	  join(', ', @Classes), $Info);

    if ($Info) {
	#
	# Retrieve just $Info
	#
	my $r = $self->RetrieveData(undef, $Args);
	if ($r == -2) {	# Hard error - don't go any further
	    return;
	}
    } else {
	#
	# Iterate through all specified classes
	#
	foreach my $class (@Classes) {
	    my $r = $self->RetrieveData($class, $Args);
	    if ($r == -2) {	# Hard error - don't go any further
		return;
	    }
	}
    }

    #
    # Restore stderr
    #
    if ($SaveSTDERR) {
	if (!open(STDERR, ">&$SaveSTDERR")) {
	    Error("Restore STDERR failed: %s", $!);
	}
    }

    if (!$Info) {
	#
	# Perform post processing of each class
	#
	my @l = $self->Class;
	foreach my $class (@l) {
	    if (exists($self->{$class}) &&
		exists($self->{$class}->{"POSTPROCESS"})) {
		$self->{$class}->POSTPROCESS;
	    }
	}
    }

    Debug("SysInfo->Retrieve: DONE");

    return $self;
}

#
# See if $Key is in $Members.
# This is similiar to ArrayFind, but uses lc() on all members of array.
#
sub FindMember {
    my $Key = shift;
    my @Members = @_;

    foreach my $k (@Members) {
	if (lc($Key) eq lc($k)) {
	    return $k;
	}
    }

    return undef;
}

#
# Retrieve $Key and return the object its found in.
# Returns mcStatus object
#
sub RetrieveKey {
    my $self = shift;
    my $Key = shift;
    my $Args = shift;
    my $Status = mcStatus->New;

    Debug("Retrieving SysInfo key <%s> ...", $Key);

    #
    # Look through all members of $self for an object which has a member
    # called $Key (case insensitive).  When we find such a class, then
    # retrieve that class data and return $Status storing the found object
    # in $Status->Result.
    #
    foreach my $cname (keys %$self) {
	my $obj = $self->{$cname};
	#
	# Look to see if this hash member is a hash itself and if so
	# if it has a Members member.
	#
	if ($obj !~ /HASH/ || !exists($obj->{Members})) {
	    next;
	}
	#
	# Get list of all member names for $obj
	#
	my $Members = $obj->Members($Args);
	if (!$Members) {
	    next;
	}

	#
	# Look through Members for a matching member name so we can get the
	# class name.
	#
	my $FoundClass;
	foreach my $mname (sort keys %$Members) {
	    if (Eq($Members->{$mname}->Key, $Key)) {
		$FoundClass = $Members->{$mname}->ClassName;
		Debug("mcSysInfo->RetrieveKey: For <%s> found=<%s> class=<%s>",
		      $Key, $mname, $FoundClass);
		last;
	    }
	}
	if (!$FoundClass) {
	    next;
	}

	#
	# We found the right object.
	# Retrieve everything in this class.
	# XXX For optimum performance we should only retrieve what we
	# were asked too.
	#
	$self->Class($FoundClass);
	if ($self->Retrieve) {
	    # Success
	    $Status->Code($Status->SUCCESS);
	    # Save the object as the Result for the caller
	    $Status->Result($obj);
	    return $Status;
	} else {
	    # Failure
	    $Status->Code($Status->FAILURE);
	    $Status->ErrMsg("Retrieve data for class %s failed.", $cname);
	    return $Status;
	}
    }

    # Not found
    $Status->Code($Status->FAILURE);
    $Status->ErrMsg("No such mcSysInfo key as \"%s\"", $Key);
    return $Status;
}

1; # Make require/use happy
