#
# $Id: mcOptions.pm,v 1.19 2003/04/19 21:12:45 mcooper Exp $
#
# Command Line Option handling
#
# We accept both '-long' and '--long' long options as well as '-X' short
# options.
#
# The OptionsTable hash keys we look for are:
#	long=>name	Long option name (no -)
#	short=>X	Short option char (no -)
#	bool=>1		Boolean value, requires arg YES/NO/TRUE/FALSE
#	argreq=>1	Argument is required
#	argopt=>1	Argument is optional
#	argnames=>[]	Array of valid arguments
#	hidden=>1	Don't show during Usage
#	envvar=>name	Look for $ENV{$name} if not found on command line
#	default=>value	Default value to set prior to CLI parsing
#	RC=>1		Check in system and user RC for this value
# Returned keys:
#	ENABLED=>1	If this option was specified by command line
#	ARG=>string	Argument string if argopt or argreq
#	ARGARRAY=>[]	Array value of ARG
#

package mcOptions;

use mcDebug;

require 5.6.1;
require Exporter;

BEGIN {
    #
    # Export function
    #
    @ISA      = qw(Exporter);

    #
    # Variables and functions being auto exported
    #
    @EXPORT   = qw(
		   OptionsParse
		   OptionsUsage
		   OptionsLong
		  );
}

#
# Our variables
#
my $OptTab;

sub Eq {
    my $s1 = shift;
    my $s2 = shift;

    return lc($s1) eq lc($s2);
}

#
# Handle option error
#
sub OptErr {
    my $Args = shift;
    my $Msg = $Args->{ErrMsg};

    if (!$Args || !$Msg) {
	return;
    }

    if (defined($Args->{ErrorFunc})) {
	$Args->{ErrorFunc}($Msg);
    }
}

#
# Handle option error
#
sub OptError {
    my $Args = shift;

    OptErr($Args);
    OptionsUsage($Args);
}

#
# Is $Name an option?
#
sub IsOpt {
    my $Name = shift;

    if ($Name =~ /^-/ || $Name =~ /^\+/) {
	return 1;
    } else {
	return 0;
    }
}

#
# Set a boolean value
# Returns undef on success
# Returns scalare with err on failure
#
sub SetBoolean {
    my $Opt = shift;
    my $Value = shift;

    if (!length($Value)) {
	return 'No boolean value was specified';
    }
    if (!$Opt) {
	return 'Missing parameters passed to SetBoolean';
    }

    if (Eq($Value, 'yes') || Eq($Value, 'true')) {
	$Opt->{ENABLED} = 1;
    } elsif (Eq($Value, 'no') || Eq($Value, 'false')) {
	$Opt->{ENABLED} = 0;
    } else {
	# Failed
	return sprintf("Boolean value \"%s\" was not recognized", $Value);
    }

    # Success
    return;
}

#
# Return the option entry from OptionsTable for OptionName
#
sub GetOpt {
    my $Args = shift;
    my $Tab = $Args->{OptionsTable};
    my $Name = $Args->{OptionName};

    foreach my $key (keys %$Tab) {
	my $Entry = $Tab->{$key};

	if (Eq($Name, $key)) {
	    return $Entry;
	}

	# Check short option
	if ($Name eq '-' . $Entry->{short}) {
	    return $Entry;
	}

	# Check all long options
	my @Longs;
	if (ref($Entry->{long}) eq 'ARRAY') {
	    @Longs = @{ $Entry->{long} };
	} else {
	    # Must be a scalar
	    push(@Longs, $Entry->{long});
	}

	foreach my $long (@Longs) {
	    #
	    # Check for $long --$long -$long
	    #
	    if ($Name eq $long ||
		($Name eq '--' . $long) ||
		($Name eq '-' . $long)) {
		return $Entry;
	    }
	}
    }
}

#
# Return the long option name for $OptName
#
sub OptionsLong {
    my $Args = shift;
    my $OptName = $Args->{OptName};
    my $Tab = $Args->{OptionsTable};

    if (!$Args || !$OptName || !$Tab) {
	return;
    }

    my $o = $Tab->{$OptName};
    if (!$o) {
	return;
    }

    if (ref($o->{long}) eq 'ARRAY') {
	return '--' . $o->{long}[0];
    } else {
	return '--' . $o->{long};
    }
}

#
# Set Option
#
sub SetOption {
    my $Args = shift;
    my $Opt = $Args->{Option};
    my $Value = $Args->{Value};
    my $LineNo = $Args->{LineNo};
    my $File = $Args->{File};
    my $Status = 0;

    if ($Opt->{argreq} || $Opt->{argopt}) {
	#
	# Set $Value as the value for this option
	#
	$Opt->{ARG} = $Value;
	# Check validity
	$Args->{Option} = $Opt;
	$Args->{Silence} = 1;
	if (IsValid($Args) < 0) {
	    my $err = $Args->{ErrMsg};	# From IsValid()
	    if ($File && $LineNo) {
		$Args->{ErrMsg} = sprintf("%s: line %d: warning: %s",
					  $File, $LineNo, $err);
	    } else {
		$Args->{ErrMsg} = $err;
	    }
	    OptErr($Args);
	    $Status = -1;
	}
	delete($Args->{Option});
	delete($Args->{Silence});
    } else {
	#
	# This option is a boolean
	#
	if (Eq($Value, 'yes') || Eq($Value, 'true')) {
	    $Opt->{ENABLED}++;
	} elsif (Eq($Value, 'no') || Eq($Value, 'false')) {
	    $Opt->{ENABLED} = 0;
	} else {
	    if ($File && $LineNo) {
		$Args->{ErrMsg} = sprintf("%s: line %d: warning: Boolean <%s> is invalid.",
					  $File, $LineNo, $Value);
	    } else {
		$Args->{ErrMsg} = $err;
	    }
	    OptErr($Args);
	    $Status = -1;
	}
    }

    return $Status;
}

#
# Set command using $Line
#
sub Set {
    my $Args = shift;
    my $Line = $Args->{Line};
    my $LineNo = $Args->{LineNo};
    my $Tab = $Args->{OptionsTable};
    my $File = $Args->{File};

    #
    # $Line should be key=value
    #
    my @p = split('=', $Line);
    my $key = lc($p[0]);
    my $value = $p[1];

    Debug("ParseRC:\t\t <%s>=<%s>", $key, $value);

    if (!length($key)) {
	$Args->{ErrMsg} = sprintf("%s: line %d: warning: Option <%s> is invalid.",
				  $File, $LineNo, $key);
	if ($File && $LineNo) {
	    OptErr($Args);
	} else {
	    Debug($Args->{ErrMsg});
	}
	return -1;
    }
	
    my $opt = GetOpt({OptionName=>$key, OptionsTable=>$Tab});
    if (!$opt) {
	$Args->{ErrMsg} = sprintf("%s: line %d: warning: Option <%s> is invalid.",
				  $File, $LineNo, $key);
	if ($File && $LineNo) {
	    OptErr($Args);
	} else {
	    Debug($Args->{ErrMsg});
	}
	return -1;
    }

    return SetOption({Option=>$opt, Value=>$value,
		      File=>$File, LineNo=>$LineNo});
}

#
# Parse the options RC $File.
#
sub ParseRC {
    my $Args = shift;
    my $Tab = $Args->{OptionsTable};
    my $File = $Args->{File};

    if (!$Tab) {
	return -1;
    }

    if (!(-f $File)) {
	Debug("ParseRC: %s does not exist.", $File);
	return 0;
    }

    Debug("ParseRC: Reading <%s>", $File);

    if (!open($FH, '<', $File)) {
	$Args->{ErrMsg} = sprintf("%s: Open for read failed: %s", $File, $!);
	OptError($Args);
	return -1;
    }

    my $LineNo = 0;
    while (my $Line = <$FH>) {
	++$LineNo;
	if ($Line =~ /^\#/) {
	    next;
	}
	chomp($Line);

	#
	# Looking for: command $cmdarg
	#
	Debug("ParseRC:\t Read Line=<%s>", $Line);

	$Args->{Line} = $Line;
	if ($Line =~ /=/) {
	    Set($Args);
	} else {
	    Debug("ParseRC:\t %s: line %d: Command <%s> is not recognized.",
		  $File, $LineNo, $Line);
	}
	delete($Args->{Line});
    }

    close($FH);

    return 0;
}

#
# Parse the system and user RC files.
# Any option in OptionsTable which has RC=>1 is checked.
#
sub OptionsParseRC {
    my $Args = shift;

    if ($Args->{SystemRC}) {
	$Args->{File} = $Args->{SystemRC};
	ParseRC($Args);
    }
    if (my $rc = $ENV{"SYSINFO_RC"}) {
	$Args->{File} = $rc;
	ParseRC($Args);
    } elsif ($Args->{UserRC}) {
	$Args->{File} = $ENV{"HOME"} . '/' . $Args->{UserRC};
	ParseRC($Args);
    }
    delete($Args->{File});
}

#
# Check validity
#
sub IsValid {
    my $Args = shift;
    my $Opt = $Args->{Option};

    if ($Opt->{ARG}) {
	# Create handy array version of ARG
	my @sa = split(',', $Opt->{ARG});
	$Opt->{ARGARRAY} = \@sa;
    }

    if ($Opt->{argnames} && $Opt->{ARGARRAY}) {
	# Validate that each arg specified on CLI is valid
	my @ValidNames = @{ $Opt->{argnames} };
	foreach my $a (@{ $Opt->{ARGARRAY} }) {
	    my $Found = 0;
	    for (my $x = 0; $x <= $#ValidNames; ++$x) {
		if (lc($ValidNames[$x]) eq lc($a)) {
		    $Found = 1;
		}
	    }

	    if (!$Found) {
		$Args->{ErrMsg} = sprintf("Argument \"%s\" (used with \"%s\") is not valid.\n\t Must be one of: %s",
					  $a,
					  '--' . $Opt->{long},
					  join(',', @ValidNames));
		if (!$Args->{Silence}) {
		    OptError($Args);
		}
		return -1;
	    }
	}
    }
    # Mark it enabled
    $Opt->{ENABLED}++;

    return 0;
}

#
# Look in OptionsTable for any option with "default" and set that
# value.
#
sub SetDefaults {
    my $Args = shift;
    my $Tab = $Args->{OptionsTable};
    my $Status = 0;

    foreach my $key (keys %$Tab) {
	my $opt = $Tab->{$key};
	if (!exists($opt->{default})) {
	    next;
	}

	if (SetOption({Option=>$opt, Value=>$opt->{default}}) < 0) {
	    --$Status;
	}
    }

    return $Status;
}

#
# Parse options
#
sub OptionsParse {
    my $Args = shift;
    my $Tab = $Args->{OptionsTable};

    if (!$Tab) {
	return -1;
    }

    # Set defaults as specified in OptionsTable
    SetDefaults($Args);

    # Parse the system and user RC files first
    OptionsParseRC($Args);

    #
    # Parse command line
    #
    for (my $i = 0; $i <= $#ARGV; ++$i) {
	my $Arg = $ARGV[$i];
	my $ArgNext = $ARGV[$i+1];

	# Special check for help
	if ($Arg eq '--help' || $Arg eq '-help') {
	    OptionsUsage($Args);
	    exit(0);
	}

	if (IsOpt($Arg)) {
	    $Args->{OptionName} = $Arg;
	    my $Opt = GetOpt($Args);
	    if (!$Opt) {
		$Args->{ErrMsg} = sprintf("Option \"%s\" is invalid.", $Arg);
		OptError($Args);
		return -1;
	    }
	    # Success
	    my $CheckValid = 1;
	    if ($Opt->{bool}) {
		$CheckValid = 0;
		if (my $err = SetBoolean($Opt, $ArgNext)) {
		    $Args->{ErrMsg} = sprintf("Option \"%s\" requires a boolean argument: %s",
					      $Arg, $err);
		    OptError($Args);
		    return -1;
		}
		++$i;
	    } elsif ($Opt->{argreq}) {
		# An argument to this option is required
		if (!length($ArgNext) || IsOpt($ArgNext)) {
		    $Args->{ErrMsg} = sprintf("Option \"%s\" requires an argument.",
					      $Arg);
		    OptError($Args);
		    return -1;
		}
		$Opt->{ARG} = $ArgNext;
		++$i;
	    } elsif ($Opt->{argopt}) {
		# We accept an optional argument
		if (!IsOpt($ArgNext)) {
		    $Opt->{ARG} = $ArgNext;
		    ++$i;
		}
	    }

	    if ($CheckValid) {
		# Check validity
		$Args->{Option} = $Opt;
		my $r = IsValid($Args);
		delete($Args->{Option});
		if ($r < 0) {
		    return -1;
		}
	    }
	} elsif ($Args->{WantArgs}) {
	    # Provide this argument to the caller
	    push(@{ $Args->{ARGS} }, $Arg);
	} else {
	    # Extra arg specified which is unwanted
	    $Args->{ErrMsg} = sprintf("Argument \"%s\" is invalid.", $Arg);
	    OptError($Args);
	    return -1;
	}
    }

    #
    # Check for environment vars
    #
    foreach my $key (keys %$Tab) {
	my $o = $Tab->{$key};
	#
	# If not set (by command line) and an env var name is given,
	# use that value from the environment (if any) to enable this option.
	#
	if (!$o->{ENABLED} && $o->{envvar}) {
	    my $evar = $o->{envvar};
	    if (my $val = $ENV{$evar}) {
		$o->{ENABLED} = 1;
		$o->{ARG} = $val;
	    }
	}
    }
}

#
# Sort func for options
#
sub SortOpt {
    # Sort on long options first, if not use short
    my $aopt;
    my $bopt;

    if (ref($OptTab->{$a}->{long}) eq 'ARRAY') {
	$aopt = $OptTab->{$a}->{long}[0];
    } elsif ($OptTab->{$a}->{long}) {
	$aopt = $OptTab->{$a}->{long};
    } else {
	$aopt = $OptTab->{$a}->{short};
    }
    if (ref($OptTab->{$b}->{long}) eq 'ARRAY') {
	$bopt = $OptTab->{$b}->{long}[0];
    } elsif ($OptTab->{$b}->{long}) {
	$bopt = $OptTab->{$b}->{long};
    } else {
	$bopt = $OptTab->{$b}->{short};
    }

    # Remove any leading opt chars
    $aopt =~ s/^-//;
    $aopt =~ s/^--//;
    $aopt =~ s/^\+//;
    $bopt =~ s/^-//;
    $bopt =~ s/^--//;
    $bopt =~ s/^\+//;

    return $aopt cmp $bopt;
}

#
# Print options usage
#
sub OptionsUsage {
    my $Args = shift;
    my $Tab = $Args->{OptionsTable};
    my $ProgName = $Args->{ProgramName} || $PROGRAM_NAME || $0;
    $OptTab = $Tab;

    if (!$Args || !$Tab) {
	return -1;
    }

    if (!$ProgName) {
	$ProgName = $ARGV[0];
	$ProgName =~ s/.*\///g;
    }

    printf STDERR "Usage:\t%s [ options ]\n", $ProgName;
    printf STDERR "\nwhere [ options ] are:\n";

    foreach my $key (sort SortOpt keys %$Tab) {
	my $o = $Tab->{$key};

	if ($o->{hidden}) {
	    next;
	}

	my $long;
	if (ref($o->{long}) eq 'ARRAY') {
	    $long = $o->{long}[0];
	} else {
	    $long = $o->{long};
	}

	if ($long !~ /^-/ && $long !~ /^\+/) {
	    $long = '--' . $long;
	}

	my $usage;
	if ($o->{bool}) {
	    $usage = 'YES|NO';
	} elsif ($o->{argusage}) {
	    $usage = $o->{argusage};
	}

	my $opts = sprintf("%s%s%s%s%s%s",
			   ($o->{short}) ? '-' . $o->{short} : '  ',
			   ($o->{short}) ? '|' : ' ',
			   ($long) ? $long : '',
			   ($usage) ? ' [' : ' ',
			   $usage,
			   ($usage) ? ']' : '');
	printf STDERR ("  %-25s %s\n", $opts, $o->{desc});
	if ($o->{argnames}) {
	    my @names = @{ $o->{argnames} };
	    printf STDERR ("  %-25s     VALUES: %s\n",
			   '',
			   join(',', @names));
	}
	if ($o->{default}) {
	    printf STDERR ("  %-25s     DEFAULT: %s\n",
			   '',
			   $o->{default});
	}
    }
}

1;
