#
# $Id: uiCLI.pm,v 1.42 2003/04/24 21:08:58 mcooper Exp $
#
# SysInfo CLI functions
#

package uiCLI;

require 5.003;
require Exporter;

use POSIX;

use mcDebug;
use mcl;
use uiGlobals;
use uiFormat;
use uiMsgLevel;
use uiClass;
use uiMisc;
use uiCliShowGeneral;
use uiCliShowLicense;
use uiCliShowHardware;
use uiCliShowSysConf;
use uiCliShowKernel;
use uiCliShowPartition;
use uiCliShowPrinter;
use uiCliShowSoftware;
use uiCliShowNetIf;
use uiCliShowNetwork;
use uiPresent;

BEGIN {
    #
    # Export function
    #
    @ISA      = qw(Exporter);

    #
    # Variables and functions being auto exported
    #
    @EXPORT   = qw(
		   CLI
		  );
}

#
# Retrieve sysinfo Class in report (raw) format
#
sub RetrieveReport {
    my $Args = shift;
    my $SysInfo = $Args->{SysInfo};
    my $SysInfoOpts = $Args->{SysInfoOpts};

    $SysInfo->ReportSet($SysInfoOpts);

    for ( ; ; ) {
	my $Line = $SysInfo->ReportLine;
	if (!defined($Line)) {
	    last;
	}
	printf "%s\n", $Line;
    }
}

#
# Retrieve sysinfo Class via standard mcSysInfo class
#
sub RetrieveData {
    my $Args = shift;
    my $SysInfo = $Args->{SysInfo};
    my $SysInfoOpts = $Args->{SysInfoOpts};
    my $Show = $Args->{Show};

    if ($Show) {
	#
	# The request is for one specific item called $Show
	#
	my $Status = $SysInfo->RetrieveKey($Show, $Args);
	if ($Status->Code == $Status->FAILURE) {
	    Error("Failed to retrieve \"%s\" info: %s",
		  $Show, $Status->ErrMsg);
	    return -1;
	}
    } else {
	#
	# Retrieve all the requested (or default) data
	#
	if (!$SysInfo->Retrieve($SysInfoOpts)) {
	    return -1;
	}
    }

    #
    # Check to see if we got any info back.
    # Warn/disable for any classes we failed to get data for.
    #
    my $Class = uiClass::ClassGet($SysInfo->Class);
    if (!$Class) {
	Debug("Retrieve: Cannot find class <%s>", $SysInfo->Class);
	return -1;
    }

    # Class Name
    my $cname = $Class->{Name};

    #
    # Present with the proper presentation function
    #
    if ($SysInfo->{$cname}->Entries) {
	#
	# Looks good
	#
	# Build CliArgs to pass to the Cli function
	#
	my $CliArgs = {};
	$CliArgs->{SysInfo} = $SysInfo->{$cname};
	if ($Show) {
	    $CliArgs->{Show} = $Show;
	}
	# Call the right Cli function
	if ($cname eq $CL_GENERAL) {
	    CliShowGeneral($CliArgs);
	} elsif ($cname eq $CL_LICENSE) {
	    CliShowLicense($CliArgs);
	} elsif ($cname eq $CL_SOFTWARE) {
	    CliShowSoftware($CliArgs);
	} elsif ($cname eq $CL_HARDWARE) {
	    CliShowHardware($CliArgs);
	} elsif ($cname eq $CL_SYSCONF) {
	    CliShowSysConf($CliArgs);
	} elsif ($cname eq $CL_KERNEL) {
	    CliShowKernel($CliArgs);
	} elsif ($cname eq $CL_PARTITION) {
	    CliShowPartition($CliArgs);
	} elsif ($cname eq $CL_PRINTER) {
	    CliShowPrinter($CliArgs);
	} elsif ($cname eq $CL_NETWORK) {
	    CliShowNetwork($CliArgs);
	} elsif ($cname eq $CL_NETIF) {
	    CliShowNetIf($CliArgs);
	} else {
	    Error("Class <%s> has no Show function defined.",
		  $cname);
	}
    } else {
	Debug("uiCLI->Retrieve: Probe didn't find any info for class <%s>",
	      $cname);

	if (Eq($cname, 'general')) {
	    Error("Could not retrieve 'general' information.  This usually indicates a problem with SysInfo.");
	} else {
	    Error("%s information is not available for your system.  This is usually because the OS doesn't supported the required interfaces.",
		  $cname);
	}
    }
}

#
# Retrieve sysinfo data and add it to $SysInfo->*->Entries
#
sub Retrieve {
    my $Args = shift;
    my $SysInfo = $Args->{SysInfo};
    my @SysInfoArgs = @{ $Args->{ARGV} };
    my @ShowList = @{ $Args->{ShowList} };
    my @RetrieveList;
    my $Failed = 0;

    Status("Probing for system information . . . ");

    my $Present = uiPresent->New;
    $Present->Start({SysInfo=>$SysInfo});

    if ($#ShowList >= 0) {
	#
	# Fetch only specific "Show" bits
	#
	foreach my $sname (@ShowList) {
	    $Args->{Show} = $sname;
	    if (RetrieveData($Args) < 0) {
		++$Failed;
	    }
	    delete($Args->{Show});
	}
    } else {
	#
	# Fetch list of all enabled classes
	#
	@RetrieveList = ClassListEnabled();

	if ($#RetrieveList < 0) {
	    # This shouldn't happen
	    Msg("No class data was specified and no default is present.\n");
	    Status("");
	    return;
	}

	foreach my $Class (@RetrieveList) {
	    my $ClassName = $Class->{Name};
	    Debug("CLI: Retrieving class <%s>", $ClassName);
	    # Specify what to retrieve
	    $SysInfo->Class($ClassName);

	    #
	    # Retrieve this class
	    #
	    #MeterCreate({Message=>'Probing system . . .'});
	    $Args->{SysInfo} = $SysInfo;

	    if (FormatIs($FT_REPORT)) {
		RetrieveReport($Args);
	    } else {
		RetrieveData($Args);
	    }
	}
    }

    $Present->End;

    $SysInfo->Destroy;

    if ($Failed) {
	return -1;
    } else {
	return 0;
    }
}

#
# Run the engine once and exit
#
sub RunOnce {
    system(join(' ', @_));
    exit(0);
}

#
# Main function to start the CLI
#
sub CLI {
    my $Args = shift;
    my @UserClassList = @{ $Args->{ClassList} };
    my @SysInfoArgs = @{ $Args->{ARGV} };
    my @ShowList = @{ $Args->{ShowList} };
    my @SysInfoCmd;

    SetIM($IM_CLI);

    Debug("Starting CLI");

    if (!$SysInfoPath && !$Args->{Host}) {
	Error("The \$SysInfoPath variable is not set.");
	Status("");
	return -1;
    }
    push(@SysInfoCmd, $SysInfoPath);

    #
    # Show copyright and exit
    #
    if ($Args->{Copyright}) {
	RunOnce(@SysInfoCmd, '--copyright');
    } elsif ($Args->{License}) {
	exit(mclShow($Args));
    }

    if (my $ms = mclCheck($Args)) {
	if ($ms->{Message}) {
	    printf STDERR $ms->{Message};
	    printf STDERR "\n";
	}
	if ($ms->{Status} != 0) {
	    exit(1);
	}
    }

    ClassSet($Args);

    #
    # Setup Debug values
    #
    #Debug({Debug=>$Debug, Func=>\&Msg});

    Retrieve($Args);

    #
    # Display hints
    #
    if ($Args->{Hints} && $#ShowList < 0 &&
	FormatIs($FT_TREE, $FT_COLUMNS) &&
	MsgLevelIs($ML_ALL, $ML_GENERAL)) {
	#
	# Display hints
	#
	my @Hints;
	if ($#UserClassList < 0) {
	    push(@Hints,
		 "HINT: Specify `--class all' to get all classes of info.");
	}
	if (!MsgLevelIs($ML_ALL)) {
	    push(@Hints,
		 "HINT: Specify `--msglevel all' to get lots of details.");
	}
	if ((InList($CL_ALL, @UserClassList) ||
	     InList($CL_SOFTWARE, @UserClassList)) &&
	    MsgLevelIs($ML_ALL) && !$Args->{SwFiles}) {
	    push(@Hints,
		 "HINT: Specify `--swfiles' to get software file info.");
	}
	if ($#Hints >= 0) {
	    printf STDERR "\n";
	    foreach my $h (@Hints) {
		printf STDERR "%s\n", $h;
	    }
	}
    }
}

1;
