#
# $Id: uiClass.pm,v 1.18 2003/05/04 00:56:58 mcooper Exp $
#
# UI Class functions
#

package uiClass;

require 5.003;

require Exporter;

use mcDebug;
use uiMisc;
use uiGlobals;
use uiFormat;
use uiPath;
use uiRetrieve;

my $Order = 0;

#
# $Class definetion
#
# Attributes members are as follows:
# 	Table->{$Arrange} - Main hash table containing path data
#			  organized by $Arrange. The key is a pathname.
#	Entries		- Object entries from mcSysInfo
#	Obj		- The object.  i.e. mcDevInfo
#	State		- The state of the entry. i.e. $PS_*
#	Summary		- Text summary describing entry value.
#	Desc		- Text describing what the entry is.
#	Value		- The value of the entry.
#	Widget		- The current display widget
#	ArrangeActive	- Currently active Arrange type
#	ArrangeSelected - The type requested by user - may lag behind Arrange
#	$widname	- Hash pointing to a widget.  One per type (HList...)
#	$WidgetName->{DataDisplayed}
#			- Boolean which indicates the data is current shown
#			  in the widget.
#

BEGIN {
    #
    # Export function
    #
    @ISA      = qw(Exporter);

    #
    # Variables and functions being auto exported
    #
    @EXPORT   = qw(
		   ClassGet ClassHideWidget ClassShow ClassNames
		   ClassImage ClassTable ClassSet ClassListEnabled
		  );
}

#
# Load modules when called
#
sub LoadModules {
    if (!$LoadModulesDone && GuiMode) {
	Debug("uiClass: Loading GUI modules ...");
	$LoadModulesDone = 1;
	eval 'use GuiWidget';
	eval 'use GuiDataDisplay';
	eval 'use GuiMeter';
	eval 'use GuiButtons';
	eval 'use GuiIcon';
	eval 'use GuiClass';
    }
}

#
# Front end to GuiHideWidget
#
sub ClassHideWidget {
    if (GuiMode) {
	return GuiClassHideWidget(@_);
    }
}

#
# Find Class
#
sub ClassGet {
    my $Name = shift;

    # If $Name is a Path, then take the first member
    if ($Name =~ /$PathSep/) {
	my @p = split($PathSep, $Name);
	$Name = $p[0];
    }

    my @List = @$Classes;
    for (my $i = 0; $i <= $#List; ++$i) {
	if (Eq($Name, $List[$i]->{Name})) {
	    return $List[$i];
	}
    }
}

#
# Return the Table for $Class of type $Arrange.
# If {Set} is provided, set the Table to {Set}.
#
sub ClassTable {
    my $Args = shift;
    my $Class = $Args->{Class};
    my $Arrange = $Args->{Arrange};
    my $Set =  $Args->{Set};

    LoadModules();

    if (!length($Arrange)) {
	$Arrange = uiArrange::ArrangeName({Class=>$Class});
    }
    if (!length($Arrange)) {
	$Arrange = $ArrangeDefault;
    }

##    Debug("ClassTable: Class=<%s> Arrange=<%s> Table=<%s> Set=<%s>",
##	  $Class->{Name}, $Arrange, $Class->{Table}->{$Arrange}, $Set);

    if ($Set) {
	# Set the table to $Set
	$Class->{Table}->{$Arrange} = $Set;
    } else {
	# Return current table
	return $Class->{Table}->{$Arrange};
    }
}

#
# Show Class $Class
#
sub ClassShowClass {
    if (GuiMode) {
	GuiClassShowClass(@_);
    }
}

#
# Show Class
#
sub ClassShow {
    my $Args = shift;

    LoadModules();

    if ($Args->{Class}) {
	$Args->{OneClass} = 1;
	ClassShowClass($Args);
    } else {
	foreach my $Class (sort ClassSort @$Classes) {
	    $Args->{Class} = $Class;
	    ClassShowClass($Args);
	}
    }

    WidgetTabSetCurrent();
    ButtonStates();

    if (!$Args->{KeepMeter}) {
	MeterDestroy();
    }
}

#
# Get and set the image for the given class
#
sub ClassImage {
    my $Args = shift;
    my $Class = $Args->{Class};
    my $Widget = $Args->{Widget};

    LoadModules();

    my $image = $Class->{Image . $Widget};

    if (!$image) {
	if (my $icon = GetIcon({Path=>$Class->{Name}})) {
	    #
	    # Store into $Class based on the $Widget name since
	    # the image is only valid on the same $Widget it was created.
	    #
	    $Class->{Image . $Widget} = $image = $Widget->Getimage($icon);
	}
    }

    return $image;
}

#
# Setup Class info
#
sub ClassSet {
    my $Args = shift;
    my @UserClassList = @{ $Args->{ClassList} };

    #
    # If the user specified a class list, then set up accordingly
    #
    if ($#UserClassList >= 0) {
	# Special hack to change 'device' -> 'hardware'
	for (my $i = 0; $i <= $#UserClassList; ++$i) {
	    if (Eq($UserClassList[$i], 'device')) {
		$UserClassList[$i] = 'hardware';
	    }
	}

	# Special case for 'all'
	if (Eq($UserClassList[0], 'all')) {
	    foreach my $Class (@$Classes) {
		$Class->{Show} = 1;
	    }
	    return;
	}

	# Initialize everything to no Show
	foreach my $Class (@$Classes) {
	    $Class->{Show} = 0;
	}

	#
	# Set classes per user instructions
	# We must be careful to preserve the order the user specified.
	#
	foreach my $name (@UserClassList) {
	    my $Class = ClassGet($name);
	    if (!$Class) {
		next;
	    }
	    $Class->{Show} = 1;
	    $Class->{Order} = $Order++;
	}
    } elsif (CliMode()) {
	# In CLI mode, set HARDWARE to Show by default
	if (my $cl = ClassGet($CL_HARDWARE)) {
	    $cl->{Show} = 1;
	    $Class->{Order} = $Order++;
	}
    }
}

#
# Return an array of class names
#
sub ClassNames {
    my @List;

    foreach my $Class (@$Classes) {
	push(@List, $Class->{Name});
    }

    return @List;
}

#
# Sort Class list
#
sub ClassSort {
    return $a->{Order} <=> $b->{Order};
}

#
# Return a list of enabled classes
#
sub ClassListEnabled {
    my @Enabled;

    my @list = sort ClassSort @$Classes;
    foreach my $Class (@list) {
	if ($Class->{Show} && !$Class->{Entries}) {
	    push(@Enabled, $Class);
	}
    }

    return @Enabled;
}

1;
