#
# $Id: uiCliShowHardware.pm,v 1.22 2003/05/07 18:08:19 mcooper Exp $
#
# SysInfo CLI Show Hardware functions
#

package uiCliShowHardware;

require 5.003;
require Exporter;

use POSIX;

use CGI qw(:all);

use mcDebug;
use uiGlobals;
use uiClass;
use uiMisc;
use uiEncode;
use uiFormat;
use uiPresent;
use uiPresentItem;
use uiPresentProp;
use uiPresentHtml;
use uiText;
use uiTextTree;
use uiTextColumns;

my $AddHistory;
my $ColumnFmt = undef;
my $BorderEnable = 0;

BEGIN {
    #
    # Export function
    #
    @ISA      = qw(Exporter);

    #
    # Variables and functions being auto exported
    #
    @EXPORT   = qw(
		   CliShowHardware
		  );
}

#
# Get NetIf Properties
#
sub GetNetIfProps {
    my $NetIf = shift;
    my @Props;

    my @Names = keys %$NetIf;
    foreach my $name (@Names) {
	if (ref($NetIf->{$name})) {
	    # Skip non-scalar values
	    next;
	}
	if (my $desc = $NetIf->Desc($name)) {
	    if (my $p = CreateProp($name, $desc, $NetIf->{$name})) {
		push(@Props, $p);
	    }
	}
    }

    return @Props;
}

#
# Get CpuInfo Properties
#
sub GetCpuInfoProps {
    my $CpuInfo = shift;
    my @Props;

    my @Names = keys %$CpuInfo;
    foreach my $name (@Names) {
	if (ref($CpuInfo->{$name})) {
	    # Skip non-scalar values
	    next;
	}
	if (my $desc = $CpuInfo->Desc($name)) {
	    if (my $p = CreateProp($name, $desc, $CpuInfo->{$name})) {
		push(@Props, $p);
	    }
	}
    }

    #
    # Add cache info
    #
    foreach my $cache ($CpuInfo->Caches) {
	my @Names = keys %$cache;
	foreach my $name (@Names) {
	    if (my $desc = $cache->Desc($name)) {
		if (my $p = CreateProp($name, $desc, $cache->{$name})) {
		    push(@Props, $p);
		}
	    }
	}
    }

    return @Props;
}

#
# Get PresentItem properties for $DevInfo
#
sub GetProps {
    my $DevInfo = shift;

    #
    # Setup the properties for this item.
    #
    my @PropNames = keys %$DevInfo;
    my @PrProps;
    foreach my $prop (sort {$DevInfo->ORDER($a) <=>
				$DevInfo->ORDER($b)} @PropNames) {
	my $propkey = $prop;
	my $desc = $DevInfo->Desc($propkey);
	my $value = GetValue($DevInfo->{$prop});

	if (!length($value)) {
	    next;
	}

	Debug("CliShowHardware: Handle item=<%s> Prop=<%s> Desc=<%s> Value=<%s>",
	      $item, $prop, $desc, $value);
	
	if (!$desc) {
	    next;
	}

	my $PrProp = uiPresentProp->New;
	$PrProp->Desc($desc);
	$PrProp->Key($propkey);
	$PrProp->Value($value);
	push(@PrProps, $PrProp);
    }

    if (my @DescProps = GetDescProps($DevInfo->DescList)) {
	push(@PrProps, @DescProps);
    }

    if ($DevInfo->DevSpec) {
	Debug("DevSpec for <%s> is <%s>", $DevInfo->Name, $DevInfo->DevSpec);
	if ($DevInfo->DevSpec =~ /mcCpuInfo/) {
	    if (my @Props = GetCpuInfoProps($DevInfo->DevSpec)) {
		push(@PrProps, @Props);
	    }
	} elsif ($DevInfo->DevSpec =~ /mcNetIf/) {
	    if (my @Props = GetNetIfProps($DevInfo->DevSpec)) {
		push(@PrProps, @Props);
	    }
	}
    }
    #
    # Partitions should be in DevSpec, but predate DevSpec.
    #
    if ($DevInfo->Partitions) {
	my $Prop = uiPresentProp->New;
	$Prop->OpaqueArray($DevInfo->Partitions);
	push(@PrProps, $Prop);
    }

    return @PrProps;
}

#
# Make a PresentItem
#
sub MakeItem {
    my $DevInfo = shift;
    my $Class = shift;
    my $Master = shift;

    my $PrItem = uiPresentItem->New;
    $PrItem->Name($DevInfo->Name);
    $PrItem->SysInfoObj($DevInfo);
    if (exists($DevInfo->{"SUMMARY"})) {
	$PrItem->Summary($DevInfo->SUMMARY({Encode=>EncodeName}));
    }
    # Set our master (parent)
    $PrItem->Master($Master);

    if ($DevInfo->Name) {
	#
	# Add properties
	#
	if (my @PrProps = GetProps($DevInfo)) {
	    $PrItem->Props(@PrProps);
	}
    }

    return $PrItem;
}

#
# Show a DevInfo object recursively as TEXT
# --RECURSE--
#
sub ShowTextRecurse {
    my $DevInfo = shift;
    my $Class = shift;
    my $Level = shift;
    my $Master = shift;

    if (!$DevInfo) {
	return;
    }

    Debug("uiCliShowHardware->ShowTextRecurse: Node=<%s>",
	  $DevInfo->Name);

    my $pr = uiPresent->New;
    $pr->Class($Class);

    my $PrItem = $DevInfo->{"PresentItem"} ||
	MakeItem($DevInfo, $Class, $Master);

    if (EncodeIs($EN_TEXT)) {
	if (FormatIs($FT_TREE)) {
	    ShowTextTreeItem($pr, $Level, $PrItem, $MaxLen);
	} elsif (FormatIs($FT_COLUMNS)) {
	    if (!$ColumnFmt) {
		$ColumnFmt = ShowTextColumnsHeader({Class=>$Class});
	    }
	    ShowTextColumnsItem({Item=>$PrItem, Class=>$Class,
				 Fmt=>$ColumnFmt});
	} else {
	    Error("Format <%s> is unknown and unsupported [HW1].", FormatName);
	    exit(1);
	}
    }

    #
    # Show child nodes
    #
    if ($DevInfo->SlaveNodes) {
	my @children = $DevInfo->SlaveNodes;

	if (Debug) {
	    Debug("    ShowRecurse Slaves:");
	    foreach my $di (sort {$a->Name cmp $b->Name} @children) {
		Debug("\tSlave Node <%s>", $di->Name);
	    }
	}

	foreach my $child (@children) {
	    ShowTextRecurse($child, $Class, $Level + 1, $DevInfo);
	}
    }
}

#
# Show a DevInfo object recursively as HTML
# --RECURSE--
#
sub ShowHtmlRecurse {
    my $Args = shift;
    my $DevInfo = $Args->{DevInfo};
    my $Class = $Args->{Class};
    my $Level = $Args->{Level};
    my $ColCount = $Args->{ColCount};
    my $Master = $Args->{Master};
    my @Rows;

    if (!$DevInfo) {
	return;
    }

    Debug("uiCliShowHardware->ShowHtmlRecurse: Node=<%s>",
	  $DevInfo->Name);

    my $pr = uiPresent->New;
    $pr->Class($Class);

    my $PrItem = $DevInfo->{"PresentItem"} ||
	MakeItem($DevInfo, $Class, $Master);

    if (FormatIs($FT_TREE)) {
	my @r = HtmlTreeItem({Present=>$pr, Level=>$Level,
				 Item=>$PrItem});
	if ($#r >= 0) {
	    push(@Rows, @r);
	}
    } elsif (FormatIs($FT_COLUMNS)) {
	my @r = HtmlColumnItem({Item=>$PrItem, Class=>$Class});
	if ($#r >= 0) {
	    push(@Rows, @r);
	}
    } else {
	Error("Format <%s> is unknown and unsupported [HW2].", FormatName);
	exit(1);
    }

    #
    # Show child nodes
    #
    if ($DevInfo->SlaveNodes) {
	my @children = $DevInfo->SlaveNodes;

	if (Debug) {
	    Debug("    ShowRecurse Slaves:");
	    foreach my $di (sort {$a->Name cmp $b->Name} @children) {
		Debug("\tSlave Node <%s>", $di->Name);
	    }
	}
	
	my @ChildRows;
	foreach my $child (@children) {
	    my @rows = ShowHtmlRecurse({DevInfo=>$child, Class=>$Class,
					Level=>$Level + 1, Master=>$DevInfo});
	    if ($#rows >= 0) {
		push(@ChildRows, @rows);
	    }
	}
	if ($#ChildRows >= 0) {
	    if (FormatIs($FT_COLUMNS)) {
		push(@Rows, @ChildRows);
	    } else {
		push(@Rows, Tr(td({colspan=>$ColCount},
				  div({style=>"margin-left: .5in"},
				      table({border=>$BorderEnable,
					     width=>'100%',
					     cellpadding=>0,
					     cellspacing=>0},
					    @ChildRows)))));
	    }
	}
    }

    return @Rows;
}

#
# Show HTML
#
sub HardwareShowHtml {
    my $Args = shift;
    my $Class = $Args->{Class};
    my @Nodes = @{ $Args->{Nodes} };
    my @Rows;

    if (FormatIs($FT_COLUMNS)) {
	if (my $r = HtmlColumnHeaders($Class)) {
	    push(@Rows, $r);
	}
    }

    my $ColCnt = HtmlColumnCount;

    #
    # Get the rows for each Node
    #
    foreach my $child (@Nodes) {
	my @rows = ShowHtmlRecurse({DevInfo=>$child, Class=>$Class, Level=>0,
				    ColCount=>$ColCnt});
	if ($#rows >= 0) {
	    push(@Rows, @rows);
	}
    }

    if ($#Rows >= 0) {
	print table({border=>$BorderEnable,
		     width=>'100%',
		     cellpadding=>0,
		     cellspacing=>(FormatIs($FT_COLUMNS)) ? 3 : 0},
		     @Rows), "\n";
    }
}

#
# CLI Show Hardware class info
#
sub CliShowHardware {
    my $Args = shift;
    my $SysInfo = $Args->{SysInfo};

    Debug("CLI: ShowHardware");

    if (!$SysInfo) {
	Debug("CliShowHardware: No SysInfo data was found.");
	return -1;
    }

    my $Class = ClassGet($CL_HARDWARE);
    my $Present = uiPresent->New;
    $Present->Class($Class);

    $Present->StartClass;

    #
    # Set column widths if this is $FT_COLUMNS
    # A side effect of this is that we build and maintain a
    # PresentItem object for each SysInfo object.  This increases
    # memory utilization at the cost of speed.
    #
    if (EncodeIs($EN_TEXT) && FormatIs($FT_COLUMNS)) {
	my $Entries = $SysInfo->{"Entries"};
	foreach my $name (keys %$Entries) {
	    my $node = $Entries->{$name};
	    # Save the PresentItem node for later use
	    $node->{"PresentItem"} = MakeItem($node, $Class,
					      $node->MasterNode);
	    ColumnsSetWidth({Class=>$Class, Item=>$node->{"PresentItem"}});
	}
    }

    if (EncodeIs($EN_TEXT)) {
	#
	# Show each top level slave node and its children
	#
	my @Items;
	foreach my $child ($SysInfo->Tree->SlaveNodes) {
	    ShowTextRecurse($child, $Class, 0, undef);
	}
    } elsif (EncodeIs($EN_HTML)) {
	my @Nodes = $SysInfo->Tree->SlaveNodes;
	HardwareShowHtml({Nodes=>\@Nodes, Class=>$Class});
    }

    $Present->EndClass;
}

1;
