#
# $Id: uiPresent.pm,v 1.26 2003/04/20 00:07:11 mcooper Exp $
#
# Presentation main object
#

#
# Presentation object hierarchy:
#	$Present
#		Items		Hierarchical tree of PresentItem objs
#		ItemList	Horizontal array of PresentItem objs
#		Class		A ClassInfo hash
#		Class		A ClassInfo hash
#	$Present->$Item->$PresentItem
#		Name		Name of this item
#		Summary		Summary describing item
#		Props		Array of PresentProp objs
#	$Present->$Item->$PresentItem->$Prop
#		ORDER		Order to list/show property
#		Desc		Description of this property
#		Key		Keyword for this property
#		Value		Value of property
#
package uiPresent;
require 5.003;

require Exporter;

use uiMisc;
use uiGlobals;
use uiFormat;
use uiEncode;
use uiMsgLevel;
use uiText;
use uiTextTree;
use uiTextColumns;
use uiHtml;
use uiPresentHtml;

BEGIN {
    #
    # Export function
    #
    @ISA      = qw(Exporter);

    #
    # Variables and functions being auto exported
    #
    @EXPORT   = qw(
		   GetDescProps
		  );
}

#
# Object Constructor
#
sub New {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = {};
    my $Err;

    $self->{"Banner"} = undef;
    $self->{"Class"} = undef;
    $self->{"SysInfo"} = undef;
    $self->{"Items"} = [];
    $self->{"ItemList"} = [];

    bless ($self, $class);

    return $self;
}

sub Items {
    my $self = shift;

    if (!@_) {
	return @{ $self->{"Items"} };
    }

    @{ $self->{"Items"} } = @_;

    return;
}

sub ItemList {
    my $self = shift;

    if (!@_) {
	#
	# If ItemList is not set, have MakeItemList build it
	#
	my @a = @{ $self->{"ItemList"} };
	if ($#a < 0) {
	    @a = $self->MakeItemList;
	    if ($#a >= 0) {
		# Add it for future calls
		@{ $self->{"ItemList"} } = @a;
		return @a;
	    }
	    return;
	}
	return @a;
    }

    @{ $self->{"ItemList"} } = @_;

    return;
}

sub Banner {
    my $self = shift;

    if (!@_) {
	if ($self->{"Banner"}) {
	    return $self->{"Banner"};
	} elsif (exists($self->{"Class"})) {
	    return $self->{"Class"}->{"Banner"};
	}
    }

    $self->{"Banner"} = shift;

    return;
}

sub Class {
    my $self = shift;

    if (!@_) {
	return $self->{"Class"};
    }

    $self->{"Class"} = shift;

    return;
}

sub SysInfo {
    my $self = shift;

    if (!@_) {
	return $self->{"SysInfo"};
    }

    $self->{"SysInfo"} = shift;

    return;
}

#
# Recursively find and add items to $List
# --RECURSE--
#
sub MakeItemListRecurse {
    my $Args = shift;
    my $Item = $Args->{Item};
    my $List = $Args->{List};

    if (!$Item) {
	return;
    }

    push(@$List, $Item);

    if ($Item->Slaves) {
	my @slaves = $Item->Slaves;
	MakeItemListRecurse({Item=>$slaves[0], List=>$List});
    }

    if ($Item->Next) {
	MakeItemListRecurse({Item=>$Item->Next, List=>$List});
    }
}

#
# Make ItemList from Items
#
sub MakeItemList {
    my $self = shift;
    my @Items = $self->Items;
    my @List;

    foreach my $item (@Items) {
	MakeItemListRecurse({Item=>$item, List=>\@List});
    }

    return @List;
}

#
# Get list of Description items (DescList) and return PresentProps.
#
sub GetDescProps {
    my @DescList = @_;

    my @PrProps;
    foreach my $d (@DescList) {
	my $Prop = uiPresentProp->New;
	$Prop->Key($d->{"Key"});
	$Prop->Desc($d->{"Tag"});
	$Prop->Value($d->{"Value"});
	push(@PrProps, $Prop);
    }

    return @PrProps;
}

#
# Start a presentation
#
sub Start {
    my $self = shift;
    my $Args = shift;
    my $SysInfo = $Args->{SysInfo};

    if ($SysInfo) {
	$self->SysInfo($SysInfo);
    }

    if (EncodeIs($EN_HTML)) {
	HtmlStart($self);
    }
}

#
# Start a presentation for a class
#
sub StartClass {
    my $self = shift;

    if (EncodeIs($EN_TEXT)) {
	TextStartClass($self);
    } elsif (EncodeIs($EN_HTML)) {
	HtmlStartClass($self);
    }
}

#
# End a presentation
#
sub End {
    my $self = shift;

    if (EncodeIs($EN_HTML)) {
	HtmlEnd($self);
    }
}

#
# End a class presentation
#
sub EndClass {
    my $self = shift;

    if (EncodeIs($EN_HTML)) {
	HtmlEndClass($self);
    }
}

#
# Look for $name in @array
#
sub ArrayFind {
    my $name = shift;
    my @array = @_;

    for (my $i = 0; $i <= $#array; ++$i) {
	if ($name eq $array[$i]->Name) {
	    return $array[$i];
	}
    }

    return undef;
}

#
# Show $Item recursively
# --RECURSE--
#
sub PresentTreeRecurse {
    my $Present = shift;
    my $Level = shift;
    my $Item = shift;

    if (!$Item) {
	return;
    }

    # If $Item is just a name, then look it up and make it right
    if ($Item !~ /=HASH/) {
	my $item = ArrayFind($Item, $Present->Items);
	if (!$item) {
	    Debug("PresentTree: Cannot find item named <%s>", $item);
	    return;
	}
	$Item = $item;
    }

    if (EncodeIs($EN_TEXT)) {
	ShowTextTreeItem($Present, $Level, $Item);
    }

    if ($Item->Slaves) {
	my @slaves = $Item->Slaves;
	PresentTreeRecurse($Present, $Level+1, $slaves[0]);
    }

    if ($Item->Next) {
	PresentTreeRecurse($Present, $Level, $Item->Next);
    }
}

#
# Show Tree
#
sub PresentTreeText {
    my $Args = shift;
    my $Present = $Args->{uiPresent};
    my $Level = 0;

    my @Items = $Present->Items;
    my $f = $Items[0];
    if ((exists($f->{Next}) && $f->Next) ||
	(exists($f->{Slaves}) && $f->Slaves)) {
	#
	# Recursive iteration
	#
	foreach my $item (@Items) {
	    PresentTreeRecurse($Present, $Level, $item);
	}
    } else {
	#
	# Do non-recursive array iteration
	#
	my $MaxLen = 0;
	foreach my $item ($Present->Items) {
	    my $len = length($item->Desc);
	    if (MsgLevelIs($ML_ALL)) {
		$len = $len + length($item->Name) + 3;
	    }
	    if ($len > $MaxLen) {
		$MaxLen = $len;
	    }
	}
	foreach my $item ($Present->Items) {
	    ShowTextTreeItem($Present, $Level, $item, $MaxLen);
	}
    }
}

#
# Show presentation
#
sub Show {
    my $self = shift;
    my $Args = shift;

    $self->StartClass;

    $Args->{uiPresent} = $self;

    if (FormatIs($FT_TREE)) {
	if (EncodeIs($EN_TEXT)) {
	    PresentTreeText($Args);
	} elsif (EncodeIs($EN_HTML)) {
	    PresentHtml($Args);
	}
    } elsif (FormatIs($FT_COLUMNS)) {
	if (EncodeIs($EN_TEXT)) {
	    ShowTextColumns($Args);
	} elsif (EncodeIs($EN_HTML)) {
	    PresentHtml($Args);
	}
    } else {
	Error("Format <%s> is not known and is unsupported.", FormatName);
	exit(1);
    }

    $self->EndClass;
}

1;
