#
# $Id: uiTextColumns.pm,v 1.13 2003/05/06 18:46:11 mcooper Exp $
#
# Columns object
#

package uiTextColumns;
require 5.003;

require Exporter;

use Term::ReadKey;	# For GetTerminalSize

use mcDebug;
use mcGetSize;
use uiMisc;
use uiGlobals;
use uiMsgLevel;
use uiTextProps;
use uiCliMsg;

BEGIN {
    #
    # Export function
    #
    @ISA      = qw(Exporter);

    #
    # Variables and functions being auto exported
    #
    @EXPORT   = qw(
		   ShowTextColumns ShowTextColumnsItem
		   ShowTextColumnsHeader ColumnsSetWidth
		  );
}

#
# Return width of the user's terminal
#
sub TermWidth {
    my $Default = 80;

    #
    # Workaround so GetTerminalSize doesn't die() when STDOUT is a file
    #
    $ENV{'COLUMNS'} = 80;
    $ENV{'LINES'} = 24;

    my @Term = GetTerminalSize(STDOUT);
    if ($#Term >= 0) {
	Debug("GetTerminalSize returned: %s", join(',', @Term));
	return $Term[0] || $Default;
    }

    Debug("GetTerminalSize failed to automatically determine terminal width");

    return $Default;
}

#
# Using an array of percentages, convert that into a printf() format string
# which uses no more than the terminal width.
#
sub GetFmt {
    my @Headers = @_;

    my $MaxWidth = TermWidth();

    # Check to see if we're going to enforce minimums.
    # If any member of Minimums is == 0, then we don't enforce.
    my $EnforceMin = 1;
    foreach my $hdr (@Headers) {
	if (!Eq($hdr->{Type}, 'text')) {
	    next;	# Skip non text headers
	}
	if ($hdr->{MinWidth} == 0) {
	    $EnforceMin = 0;
	    last;
	}
    }

    my @Fmt;
    my $Used = 0;
    foreach my $hdr (@Headers) {
	if (!Eq($hdr->{Type}, 'text')) {
	    next;	# Skip non text headers
	}

	my $Min = $hdr->{MinWidth};
	if (!exists($hdr->{MinWidth}) ||
	    ($hdr->{MinWidth} == -1 && $hdr->{MinOver})) {
	    $Min = $hdr->{MinOver};
	}
	# Reset -1 to 0
	if ($Min < 0) {
	    $Min = 0;
	}

	my $w = $Min;
	if (my $pct = $hdr->{WidthPct}) {
	    # Calculate width based on percentage
	    $w = ($pct / 100) * $MaxWidth;
	    # Deduct 2 for rounding and 1 for space between args
	    $w -= 3;
	    # Insure width is not less than the minimum
	    if ($w < $Min) {
		$w = $Min;
	    }
	}

	#
	# Insure  we don't exceed MaxWidth, but only if we're enforcing
	# minimums.
	#
	if ($EnforceMin && ($Used + $w >= $MaxWidth)) {
	    $w = $MaxWidth - $Used;
	    if ($w < 0) {
		next;
	    }
	}

	# Make sure it doesn't exceed MaxWidth if one was given
	if (exists($hdr->{MaxWidth})) {
	    # -1 means use whatever space is left before we hit term MaxWidth
	    if ($hdr->{MaxWidth} == -1) {
		$w = $MaxWidth - $Used - 1;
	    } elsif ($w > $hdr->{MaxWidth}) {
		$w = $hdr->{MaxWidth};
	    }
	}

	# Width of this fmt + SPACE
	$Used += $w + 1;

	$fmt;
	if ($w == 0) {
	    $fmt = '%s';
	} else {
	    if (Eq($hdr->{"HAlign"}, 'right')) {
		# Right align
		$fmt = sprintf("%%%d.%ds", $w, $w);
	    } else {
		# Left align
		$fmt = sprintf("%%-%d.%ds", $w, $w);
	    }
	}
	push(@Fmt, $fmt);
    }

    return @Fmt;
}

#
# Show column header in TEXT + COLUMNS format
#
sub ShowTextColumnsHeader {
    my $Args = shift;
    my $Class = $Args->{Class};
    my @Headers = @{ $Class->{Headers} };

    my @FmtArgv = GetFmt(@Headers);
    my $FmtStr = join(' ', @FmtArgv);

    #
    # Get the names for the top of each column
    #
    my @ColHdrs;
    foreach my $hdr (@Headers) {
	if (!Eq($hdr->{Type}, 'text')) {
	    next;	# Skip non text headers
	}
	push(@ColHdrs, $hdr->{Name});
    }

    # Print the column header
    CliMsg($FmtStr, @ColHdrs);
    CliMsg("\n");

    return $FmtStr;
}

#
# Show item header in TEXT + COLUMNS format
#
sub ShowTextColumnsItem {
    my $Args = shift;
    my $Item = $Args->{Item};
    my $Class = $Args->{Class};
    my $Fmt = $Args->{Fmt};
    my @Headers = @{ $Class->{Headers} };

    my @Argv;
    foreach my $hdr (@Headers) {
	if (!Eq($hdr->{Type}, 'text')) {
	    next;	# Skip non text headers
	}

	my $value;
	if (my $key = $hdr->{Key}) {
	    $value = GetItemValue($key, $Item, $hdr);
	}

	push(@Argv, (length($value)) ? $value : '-');
    }

    CliMsg($Fmt, @Argv);
    CliMsg("\n");

    if (MsgLevelIs($ML_ALL, $ML_CONFIG, $ML_DESC)) {
	ShowTextProps({Item=>$Item, Class=>$Class});
    }
}

#
# Set minimum field override values
#
sub ColumnsSetWidth {
    my $Args = shift;
    my $Item = $Args->{Item};
    my $Class = $Args->{Class};
    my @Headers = @{ $Class->{Headers} };

    foreach my $hdr (@Headers) {
	if (!Eq($hdr->{Type}, 'text')) {
	    next;	# Skip non text headers
	}
	if (my $key = $hdr->{Key}) {
	    my $value = GetItemValue($key, $Item, $hdr);

	    my $len = length($value);
	    if ($len > $hdr->{MinOver}) {
		$hdr->{MinOver} = $len;
	    }
	}

	#
	# If no MinOver has been set (probably because there are no Item
	# values), then use the column header name for length MinOver.
	#
	if (!exists($hdr->{MinOver})) {
	    $hdr->{MinOver} = length($hdr->{Name});
	}
    }
}

#
# Show main presentation in TEXT + COLUMNS format
#
sub ShowTextColumns {
    my $Args = shift;
    my $Present = $Args->{uiPresent};
    my $Class = $Args->{Class};

    #
    # Set minimum field width overrides
    #
    my @Items = $Present->ItemList;
    foreach my $item (@Items) {
	ColumnsSetWidth({Class=>$Class, Item=>$item});
    }

    #
    # Show the column header
    #
    my $fmt = ShowTextColumnsHeader({Class=>$Class});

    #
    # Show each item
    #
    foreach my $item (@Items) {
	ShowTextColumnsItem({Class=>$Class, Fmt=>$fmt, Item=>$item});
    }
}

1;
