/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/5.0/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.14 $";
#endif

/*
 * Configuration File Key=Value (CFKV) functions
 *
 * These functions manipulate simple key=value config files (.cf)
 */

#include "mcbase.h"
#include <stdio.h>
#if HAVE_STRING_H
# include <string.h>
#endif
#include <ctype.h>
#include "cfkv.h"
#include "strutil.h"

#ifndef PRTS
#define PRTS(s) ( (s && *s) ? s : "" )
#endif

/*
 * Cleanup String by removing quotes and surrounding whitespace
 */
static char *Cleanup(char *String)
{
    char *			cp;
    char *			End;

    /* Skip over leading garbage */
    for ( ; String && *String &&
	      (*String == '"' || *String == '\'' || isspace(*String)) ;
	  ++String);

    End = &String[strlen(String)];
    for (cp = End;
	 cp && *cp && (*cp == '"' || *cp == '\'' || *cp == ';' || isspace(*cp));
	 ++cp);

    if (cp != End)
	*cp = CNULL;

    return String;
}

/*
 * Parse key=value from Line and add to Handle.
 * Line looks like: key=value
 * Returns error string on error.
 */
static char *Parse(CFKVhandle_t *		Handle,
		   char *			Line,
		   int				Flags)
{
    char *			Key;
    char *			Value;
    char *			cp;
    static char			KeyAlt[512];
    static int			KeyNum;

    if (!Line || !*Line)
	return NULL;

    Key = Line;
    cp = strchr(Key, '=');
    if (!cp)
	return "Not in \"key=value\" format";
    *cp = CNULL;
    Key = Cleanup(Key);
    if (!Key)
	return NULL;
    Key = strlower(Key);
    if (!Key)
	return NULL;
    Value = Cleanup(cp + 1);
    if (!Value)
	return NULL;

    if (!Key || !Value || !strlen(Key) || !strlen(Value))
	return "Zero length key or value";

    if (!Handle->KeyTab)
	Handle->KeyTab = bjhcreate((BjhWord_t)128); /* Arbitrary start size */

    if (FLAGS_ON(Flags, CFKV_DEBUG))
	fprintf(stderr, "CFKV/Parse: Key=<%s> Value=<%s>\n",
		PRTS(Key), PRTS(Value));

    if (!bjhadd(Handle->KeyTab, (ub1 *)strdup(Key),
		strlen(Key), strdup(Value))) {
	/*
	 * The add failed likely because the key already exists.
	 * So try to create a new key with a unique number and use that.
	 */
	snprintf(KeyAlt, sizeof(KeyAlt), "%s_key%d", Key, KeyNum++);
	if (FLAGS_ON(Flags, CFKV_DEBUG))
	    fprintf(stderr, "CFKV/Parse:\t Key=<%s> Value=<%s>\n",
		    PRTS(KeyAlt), PRTS(Value));
	if (!bjhadd(Handle->KeyTab, (ub1 *)strdup(KeyAlt), strlen(KeyAlt),
		    strdup(Value))) {
	    return "Error occured adding to hash key table";
	}
    }

    return NULL;	/* Success */
}

/*
 * Read ConfigFile and return a handle
 */
extern CFKVhandle_t *CFKVread(char *	ConfigFile,
			      char *	ProgramName,
			      char **	ErrPtr,
			      int	Flags)
{
    CFKVhandle_t *		Handle;
    static char			Error[1024];
    static char			Line[8192];
    char *			cp;
    int				LineNo = 0;

    if (!ConfigFile)
	return (CFKVhandle_t *) NULL;

    Handle = (CFKVhandle_t *) xcalloc(1, sizeof(CFKVhandle_t));
    Handle->ConfigFile = ConfigFile;
    Handle->Flags = Flags;

    Handle->FH = fopen(ConfigFile, "r");
    if (!Handle->FH) {
	if (ErrPtr) {
	    snprintf(Error, sizeof(Error), "%s: open for read failed: %s",
		     ConfigFile, SYSERR);
	    *ErrPtr = Error;
	}
	return (CFKVhandle_t *) NULL;
    }

    while (fgets(Line, sizeof(Line), Handle->FH)) {
	++LineNo;
	if (*Line == '#')		/* Comment line? */
	    continue;
	if (cp = strchr(Line, '\n'))
	    *cp = CNULL;		/* Zap newline */

	if (cp = Parse(Handle, Line, Flags)) {
	    /*
	     * On error only return if ErrPtr was set.
	     */
	    if (ErrPtr) {
		snprintf(Error, sizeof(Error), "%s: line %d: %s",
			 ConfigFile, LineNo, cp);
		*ErrPtr = Error;
		return (CFKVhandle_t *) NULL;
	    }
	}
    }

    (void) fclose(Handle->FH);
    Handle->FH = NULL;

    return Handle;
}

/*
 * Return the value found for Key
 */
extern char *CFKVgetValue(CFKVhandle_t *	Handle,
			  char *		Key)
{
    static char			KeyBuf[512];
    char *			cp;

    if (!Handle || !Key)
	return (char *) NULL;

    /*
     * Use lowercase version of Key
     */
    snprintf(KeyBuf, sizeof(KeyBuf), "%s", Key);
    cp = strlower(KeyBuf);
    if (!cp)
	return (char *) NULL;

    if (bjhfind(Handle->KeyTab, (ub1 *)cp, strlen(cp)))
	return (char *) bjhstuff(Handle->KeyTab);

    return (char *) NULL;
}

/*
 * Destroy Handle
 */
extern void CFKVdestroy(CFKVhandle_t *Handle)
{
    if (!Handle)
	return;

    if (Handle->FH)
	fclose(Handle->FH);

    if (Handle->KeyTab)
	/* XXX should do a better job destroying KeyTab */
	(void) free(Handle->KeyTab);
}

/*
 * Set to the first entry in the table.
 */
extern int CFKVsetEnt(CFKVhandle_t *Handle)
{
    if (!Handle)
	return -1;

    bjhfirst(Handle->KeyTab);

    return 0;
}

/*
 * Get the next entry in the table.
 */
extern int CFKVgetEnt(CFKVhandle_t *	Handle,
		      char **		KeyPtr,
		      char **		ValuePtr)
{
    if (!Handle || !KeyPtr || !ValuePtr)
	return -1;

    if (bjhnext(Handle->KeyTab)) {
	*KeyPtr = (char *) bjhkey(Handle->KeyTab);
	*ValuePtr = (char *) bjhstuff(Handle->KeyTab);
	return 1;
    } else {
	return 0;
    }
}
