/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/5.0/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.7 $";
#endif

/*
 * Linux ProbeCtlr specific functions
 */

#include "defs.h"
#include "device.h"

/*
 * The BusMap code is how we map ctlr's to their physical bus #.
 * We normally have no means to tell what a ctlr's physical bus # is.
 * Our approach is to scan a list of all current devices (DevList) and
 * build a map based on DevList.ClassType + DevList.Bus.  We then assign
 * a ctlr's Bus # by searching BusMap for the first unassigned Bus # for
 * the ctlr's ClassType.
 * XXX This assumes that the order devices are provided to us (usually 
 * how they are listed in /proc/pci/devices) is the sequential bus # order.
 */

/*
 * Bus Entry
 */
struct _BusEntry {
    int				Bus;			/* Bus # */
    int				Assigned;		/* Entry was used */
};
typedef struct _BusEntry	BusEntry_t;
/*
 * Map of BusEntry's
 */
struct _BusMap {
    int				ClassType;		/* Class Type */
#define MAX_BUSES		128
    BusEntry_t			Buses[MAX_BUSES];	/* Bus Entries */
    struct _BusMap *		Next;
};
typedef struct _BusMap		BusMap_t;

static BusMap_t *		BusMap;

/*
 * Assign a Bus number for ClassType
 */
static int BusMapAssign(int ClassType)
{
    BusMap_t *			bm;
    register int		i;

    SImsg(SIM_DBG, "BusMapAssign: Request for Class=%d", ClassType);

    for (bm = BusMap; bm; bm = bm->Next)
	if (bm->ClassType == ClassType) {
	    for (i = 0; i < MAX_BUSES && bm->Buses[i].Bus >= 0; ++i)
		if (!bm->Buses[i].Assigned) {
		    SImsg(SIM_DBG, "BusMapAssign:\t Assigning Bus=%d",
			  bm->Buses[i].Bus);
		    bm->Buses[i].Assigned = TRUE;
		    return bm->Buses[i].Bus;
		}
	}

    SImsg(SIM_DBG, "BusMapAssign:\t Class=%d is not in map", ClassType);

    return -1;
}

/*
 * Get the BusMap for ClassType
 */
static BusMap_t *BusMapGetByClass(int ClassType)
{
    BusMap_t *			bm;
    BusMap_t *			New;
    register int		i;

    for (bm = BusMap; bm; bm = bm->Next)
	if (bm->ClassType == ClassType)
	    return bm;

    /*
     * Doesn't exist so create and add a new entry
     */
    McDebug("BusMapCreate: Adding ClassType=%d", ClassType);
    New = (BusMap_t *) xcalloc(1, sizeof(BusMap_t));
    New->ClassType = ClassType;
    /* Initialize the new entry */
    for (i = 0; i < MAX_BUSES && New->Buses[i].Bus >= 0; ++i)
	New->Buses[i].Bus = -1;

    /* Add it to the list */
    if (BusMap) {
	New->Next = BusMap;
	BusMap = New;
    } else {
	BusMap = New;
    }

    return New;
}

/*
 * Create/update BusMap
 */
static int BusMapCreate(DevList_t *DevList)
{
    DevList_t *			dl;
    BusMap_t *			bm;
    register int		i;

    SImsg(SIM_DBG, "BusMapCreate: Creating/updating Bus Map ...");

    for (dl = DevList; dl; dl = dl->Next) {
	/* Find/create map entry for this ClassType */
	bm = BusMapGetByClass(dl->ClassType);
	/* Check to see if this DevList's Bus already exists */
	for (i = 0; i < MAX_BUSES && bm->Buses[i].Bus >= 0; ++i) {
	    if (bm->Buses[i].Bus == dl->Bus)
		break;				/* Already exists */
	}
	if (bm->Buses[i].Bus != dl->Bus) {
	    /* Does not yet exist */
	    bm->Buses[i].Bus = dl->Bus;		/* Add it */
	}
    }

    if (DebugOn) {
	SImsg(SIM_DBG, "BusMapCreate: Map is now:");
	for (bm = BusMap; bm; bm = bm->Next) {
	    SImsg(SIM_DBG, "BusMapCreate:\tClass=%d", bm->ClassType);
	    for (i = 0; i < MAX_BUSES && bm->Buses[i].Bus >= 0; ++i)
		SImsg(SIM_DBG, "BusMapCreate:\t\tBus[%d]=%d",
		      i, bm->Buses[i].Bus);
	}	
    }

    return 0;
}

/*
 * Probe a controller and use Detect calls as required.
 * NOTES:
 * - Make sure *Detect() functions set DevList->ClassType.
 */
extern int ProbeCtlrDetect(DevInfo_t *		CtlrDevInfo,
			   DevInfo_t **		TreePtr,
			   char **		SearchNames)
{
    DevInfo_t		       *ChildDevInfo = NULL;
    DevData_t			Child;
    int				ClassType;
    DevList_t		       *DevList;
    register DevList_t	       *dl;

    if (!CtlrDevInfo)
	return -1;

    McDebug("ProbeCtlrDetect: Ctlr=<%s> Bus=%d Type=%d ClassType=%d",
	    PRTS(CtlrDevInfo->Name), CtlrDevInfo->Bus, CtlrDevInfo->Type,
	    CtlrDevInfo->ClassType);

#if 0
    /*
     * If no Bus# was set, then assign one
     */
    if (CtlrDevInfo->Bus < 0) {
	CtlrDevInfo->Bus = UnitAssign(CtlrDevInfo->Type,
				      CtlrDevInfo->ClassType,
				      CtlrDevInfo->Name);
	McDebug("ProbeCtlrDetect:\t Assigned Bus#=%d",
		CtlrDevInfo->Bus);
    }
#endif

    /*
     * Create DevList based on ClassType
     */
    ClassType = CtlrDevInfo->ClassType;
    switch (ClassType) {
    case CT_FIBER:
    case CT_SCSI:	DevList = ScsiDetect(CtlrDevInfo);	break;
    case CT_ATA:	DevList = AtaDetect(CtlrDevInfo);	break;
    case CT_RAID:
	if (RaidDetect(CtlrDevInfo) < 0)
	    /*
	     * If we didn't find any RAID specific devices, then fall back
	     * to checking for SCSI devices as many arrays look like normal
	     * SCSI devices.
	     */
	    DevList = ScsiDetect(CtlrDevInfo);
	break;
    default:
	SImsg(SIM_DBG,
	      "ProbeCtlrDetect:\t<%s> ClassType %d has no Detect() func.",
	      CtlrDevInfo->Name, ClassType);
	return -1;
    }

    if (DebugOn && DevList) {
	/*
	 * Show all of DevList for debugging.
	 */
	SImsg(SIM_DBG, "ProbeCltr: DevList is as follows:");
	for (dl = DevList; dl; dl = dl->Next) {
	    SImsg(SIM_DBG,
		  "ProbeCtlrDetect:\tDevice=<%s> Class=%d Bus=%d Unit=%d LUN=%d Used=%d",
		  PRTS(dl->Name),
		  dl->ClassType, dl->Bus, dl->Unit, dl->Lun,
		  dl->Used);
	}
    }

    /*
     * Create/Update map of all Bus #'s in DevList.
     */
    BusMapCreate(DevList);

    /* 
     * Ensure the Controller's Bus # is set.
     */
    if (CtlrDevInfo->Bus < 0) {
	/*
	 * Assign Ctlr's Bus # according to the sequential order of the BusMap.
	 * The BusMap accounts for skipped Bus #'s.  i.e. SCSI Bus #'s 0 and 2.
	 *
	 * XXX Assume that the order the controller card is presented to us
	 * is the numeric Bus # order.
	 *
	 * Map CT_RAID -> CT_SCSI since RAID ctlr's are usually SCSI based.
	 */
	ClassType = (CtlrDevInfo->ClassType == CT_RAID) ?
	    CT_SCSI : CtlrDevInfo->ClassType;
	CtlrDevInfo->Bus = BusMapAssign(ClassType);

	SImsg(SIM_DBG, "ProbeCtlrDetect:\t<%s> SET Bus=%d (Class=%d)",
	      CtlrDevInfo->Name, CtlrDevInfo->Bus, ClassType);
    }

    McDebug("ProbeCtlrDetect:     Check DevList for Ctlr=<%s> Bus=%d ClassType=%d",
	    PRTS(CtlrDevInfo->Name), CtlrDevInfo->Bus,
	    CtlrDevInfo->ClassType);

    for (dl = DevList; dl; dl = dl->Next) {
	SImsg(SIM_DBG,
	      "ProbeCtlrDetect:\tChecking <%s>/<%s> Master=<%s>/<%s> Bus=%d Unit=%d LUN=%d",
	      PRTS(dl->Name),
	      (dl->DevInfo) ? PRTS(dl->DevInfo->Name) : "",
	      (dl->DevInfo && dl->DevInfo->Master) ?
	      PRTS(dl->DevInfo->Master->Name) : "",
	      (dl->DevInfo && dl->DevInfo->MasterName) ?
	      PRTS(dl->DevInfo->MasterName) : "",
	      dl->Bus, dl->Unit, dl->Lun);

	/* Add the device to this Controller if it's the same Bus */
	if ((dl->DevInfo && eq(CtlrDevInfo->Name, dl->DevInfo->MasterName)) ||
	    (dl->DevInfo && !dl->DevInfo->MasterName &&
	     dl->Bus == CtlrDevInfo->Bus)) {
	    if (dl->DevInfo && !dl->DevInfo->Master)
		dl->DevInfo->Master = CtlrDevInfo;
	    if (dl->DevInfo->Bus < 0)
		dl->DevInfo->Bus = dl->Bus;

	    SImsg(SIM_DBG, "ProbeCtlrDetect:\tAdding device <%s> with parent <%s>",
		  PRTS(dl->DevInfo->Name), PRTS(dl->DevInfo->Master->Name));

	    if (DevAdd(dl->DevInfo, TreePtr, SearchNames) == 0)
		dl->Used = TRUE;
	}
    }

    if (DebugOn) {
	/*
	 * Check to see if we missed adding any devices to a ctlr.
	 * XXX Note that if a disk is attached to a ctlr which is never
	 * identified as a disk ctlr, then the disk device is orphaned in
	 * DevList and never identified.  We need to catch this in some
	 * upper level function.
	 */
	for (dl = DevList; dl; dl = dl->Next) {
	    if (dl->Used)
		continue;
	    SImsg(SIM_DBG,
		  "ProbeCtlrDetect:\tDevice <%s>/<%s> is not assigned a ctlr; Bus=%d Unit=%d LUN=%d",
		  PRTS(dl->Name),
		  (dl->DevInfo) ? PRTS(dl->DevInfo->Name) : "",
		  dl->Bus, dl->Unit, dl->Lun);
	}
    }

    return 0;
}

/*
 * Check for orphaned disks in DevList
 */
extern int DiskDevListCheck(DevInfo_t **	TreePtr,
			    char **		SearchNames,
			    DevList_t *		DevList)
{
    DevList_t		       *dl;
    static DevFind_t		Find;
    int				Status = 0;

    SImsg(SIM_DBG, "DiskDevListCheck: Checking for unassigned disks ...");

    for (dl = DevList; dl; dl = dl->Next) {
	if (dl->Used || !dl->DevInfo)
	    continue;

	/* Set Master if needed */
	if (dl->DevInfo->MasterName && !dl->DevInfo->Master) {
	    DevFindInit(&Find);
	    Find.Tree = *TreePtr;
	    Find.NodeName = dl->DevInfo->MasterName;
	    dl->DevInfo->Master = DevFind(&Find);
	}

	if (dl->DevInfo->Master) {
	    SImsg(SIM_DBG,
		  "DiskDevListCheck: Disk <%s> is unassigned; adding to <%s>",
		  PRTS(dl->DevInfo->Name), PRTS(dl->DevInfo->Master->Name));
	} else {
	    SImsg(SIM_DBG,
		  "DiskDevListCheck: Disk <%s> is unassigned; adding to top of device tree",
		  PRTS(dl->DevInfo->Name));
	}

	if (DevAdd(dl->DevInfo, TreePtr, SearchNames) == 0)
	    dl->Used = TRUE;
    }

    return Status;
}

/*
 * Check for orphaned disks
 */
extern int DiskOrphanCheck(DevInfo_t **		TreePtr,
			   char **		SearchNames)
{
    return DiskDevListCheck(TreePtr, SearchNames, ScsiDetect(NULL));
}
