/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/5.0/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.6 $";
#endif

/*
 * FreeBSD pkg_info (ports) functions.
 * These functions read the pkg_info data found in PKG_DIR directly.
 * We don't use pkg_info as its output is not very parsable.
 */

#include "defs.h"
#include <sys/types.h>
#include <sys/stat.h>

#define PKG_DIR		"/var/db/pkg"

/*
 * Set description for this SoftInfo entry
 */
static int SetDesc(SoftInfo_t *		SoftInfo,
		   char *		PkgDir)
{
    static char		File[MAXPATHLEN];
    static char		Line[8192];
    char *		cp;
    FILE *		fh;

    snprintf(File, sizeof(File), "%s/+COMMENT", PkgDir);

    McDebug("BuildSoftInfoPorts:\t Setting Desc from <%s>", File);

    fh = fopen(File, "r");
    if (!fh) {
	McDebug("BuildSoftInfoPorts:\t %s: Open for read failed: %s",
		File, SYSERR);
	return -1;
    }

    /*
     * Read just the first line (there should only be one)
     */
    if (fgets(Line, sizeof(Line), fh)) {
	(void) fclose(fh);
	if (cp = strchr(Line, '\n'))
	    *cp = CNULL;
	SoftInfo->Desc = strdup(Line);
    } else {
	McDebug("BuildSoftInfoPorts:\t <%s> Read line failed: %s",
		File, SYSERR);
	return -1;
    }

    return 0;
}

/*
 * Set verbose description for this SoftInfo entry
 */
static int SetDescVerbose(SoftInfo_t *		SoftInfo,
			  char *		PkgDir)
{
    static char		File[MAXPATHLEN];
    static char		Line[8192];
    char *		Desc = NULL;
    char *		dp = NULL;
    char *		cp;
    size_t		DescLen = 0;
    int			l;
    FILE *		fh;

    snprintf(File, sizeof(File), "%s/+DESC", PkgDir);

    McDebug("BuildSoftInfoPorts:\t Setting DescVerbose from <%s>", File);

    fh = fopen(File, "r");
    if (!fh) {
	McDebug("BuildSoftInfoPorts:\t %s: Open for read failed: %s",
		File, SYSERR);
	return -1;
    }

    /*
     * Read the contents of File into buffer Desc
     */
    while (fgets(Line, sizeof(Line), fh)) {
	if (cp = strchr(Line, '\n'))
	    *cp = CNULL;

	l = strlen(Line) + 1;			/* Add space for '\n' */
	if (!Desc) {
	    /* Create initial space */
	    dp = Desc = (char *) xmalloc(l + 1);/* Add space NULL */
	} else {
	    /* Alloc more space */
	    Desc = (char *) xrealloc(Desc, DescLen + l);/* Add space NULL */
	    /* Go to end of what we just added */
	    dp = &Desc[DescLen-1];
	}
	/* Add the line + the litteral '\n' to the buffer */
	snprintf(dp, l+1, "%s%s", Line, "\n");
	/* Advance DescLen */
	DescLen += l;
    }

    (void) fclose(fh);

    SoftInfo->DescVerbose = Desc;

    return 0;
}

/*
 * Set file info for Path into SoftInfo.
 * Returns the newly created SoftFileData_t for Path.
 */
static SoftFileData_t * SetFileInfo(char *		Path,
				    SoftInfo_t *	SoftInfo)
{
    static SoftInfo_t *		LastSoftInfo;
    static SoftFileList_t *    	LastList = NULL;
    SoftFileList_t *		NewList = NULL;
    SoftFileData_t *		Data;
    static char			p[MAXPATHLEN];
    struct stat			st;

    if (!Path || !SoftInfo)
	return (SoftFileData_t *) NULL;

    if (LastSoftInfo && LastSoftInfo != SoftInfo)
	/* Invalid LastList */
	LastList = NULL;

    /* Save for next time */
    LastSoftInfo = SoftInfo;

    Data = SoftFileDataCreate(NULL);
    if (Path[0] != '/' && SoftInfo->BaseDir) {
	snprintf(p, sizeof(p), "%s/%s", SoftInfo->BaseDir, Path);
	Data->Path = strdup(p);
    } else {
	Data->Path = strdup(Path);
    }

    /*
     * Use lstat(2) to get more details.
     */
    if (lstat(Data->Path, &st) == 0) {
	switch (st.st_mode & S_IFMT) {
	case S_IFREG: Data->Type = SFT_FILE; break;
	case S_IFDIR: Data->Type = SFT_DIR; break;
	case S_IFBLK: Data->Type = SFT_BDEV; break;
	case S_IFCHR: Data->Type = SFT_CDEV; break;
	case S_IFLNK: Data->Type = SFT_SLINK; break;
	}
	Data->FileSize = (Large_t) st.st_size;
    }

    SoftInfo->DiskUsage += Data->FileSize;

    /*
     * Create a list entry and then add to the FileList
     */
    NewList = (SoftFileList_t *) xcalloc(1, sizeof(SoftFileList_t));
    NewList->FileData = Data;
    if (!SoftInfo->FileList) {
	LastList = SoftInfo->FileList = NewList;
    } else {
	LastList->Next = NewList;
	LastList = NewList;
    }

    return Data;
}

/*
 * Read the ports +CONTENTS file and set info
 */
static int SetContents(SoftInfo_t *	SoftInfo,
		       char *		PkgDir)
{
    static char		File[MAXPATHLEN];
    static char		Line[8192];
    char *		cp;
    char *		InstalledDir = NULL;
    char *		MD5 = NULL;
    char **		Argv;
    FILE *		fh;
    int			LineNo = 0;
    int			LastFileDataExpires = 0;
    SoftFileData_t *	LastFileData = NULL;

    snprintf(File, sizeof(File), "%s/+CONTENTS", PkgDir);

    McDebug("BuildSoftInfoPorts:\t Setting contents from <%s>", File);

    fh = fopen(File, "r");
    if (!fh) {
	McDebug("BuildSoftInfoPorts:\t %s: Open for read failed: %s",
		File, SYSERR);
	return -1;
    }

    while (fgets(Line, sizeof(Line), fh)) {
	++LineNo;
	if (cp = strchr(Line, '\n'))
	    *cp = CNULL;

	/* Expire the LastFileData entry if its time */
	if (LastFileData && LineNo >= LastFileDataExpires)
	    LastFileData = NULL;

	if (Line[0] == '@') {
	    /* This is a command line */
	    Argv = Split(" ", Line+1);
	    if (!Argv) {
		McDebug("BuildSoftInfoPorts:\t %s: line %d: failed to read commands",
			File, LineNo);
		continue;
	    }

	    McDebug("BuildSoftInfoPorts:\t Command=<%s> Line=<%s>",
		    Argv[0], Line);

	    if ((EQ(Argv[0], "cwd") || EQ(Argv[0], "cd")) && Argv[1]) {
		/*
		 * There may be multiple lines like this so the first
		 * one is taken as the canonical BaseDir
		 */
		if (!InstalledDir)
		    InstalledDir = strdup(Argv[1]);
		else
		    InstalledDir = Argv[1];
		if (!SoftInfo->BaseDir)
		    SoftInfo->BaseDir = InstalledDir;
	    } else if (EQ(Argv[0], "comment") && EQN(Argv[1], "MD5:", 4)) {
		/*
		 * This line contains the MD5 sum for the file on the 
		 * previous line.
		 */
		if (LastFileData)
		    LastFileData->MD5 = strdup(Argv[1] + 4);
	    } else if (EQ(Argv[0], "pkgdep") && Argv[1]) {
		McDebug("BuildSoftInfoPorts:\t Adding pkgdep=<%s>", Argv[1]);
		StrArrayAdd(&(SoftInfo->PkgDeps), strdup(Argv[1]));
	    }

	    DestroyArgv(&Argv, -1);
	} else if (Line[0] == '+') {
	    /* This line names additional pkgs files in PkgDir */
	    McDebug("BuildSoftInfoPorts:\t PKGFILE=<%s>", Line);
	} else {
	    /* This line contains a path to an installed file */
	    LastFileData = SetFileInfo(Line, SoftInfo);
	    /* This data is only valid until the line after next */
	    LastFileDataExpires = LineNo + 2;
	}
    }

    (void) fclose(fh);

    return 0;
}

/*
 * Set ports (pkg) info
 */
static int SetPorts(char *		PkgDir,
		    char *		Pkg,
		    char *		Version,
		    SoftInfo_t **	SoftInfoTree)
{
    SoftInfo_t *	SoftInfo;

    McDebug("BuildSoftInfoPorts:\t Setting Name=<%s> Version=<%s>",
	    PRTS(Pkg), PRTS(Version));

    SoftInfo = SoftInfoCreate(NULL);
    SoftInfo->Name = strdup(Pkg);
    SoftInfo->Version = strdup(Version);

    SetDesc(SoftInfo, PkgDir);
    SetDescVerbose(SoftInfo, PkgDir);
    SetContents(SoftInfo, PkgDir);

    /* Add it to SoftInfoTree */
    SoftInfoAdd(SoftInfo, SoftInfoTree, NULL);

    return 0;
}

/*
 * Build Software Information tree
 */
extern Opaque_t BuildSoftInfoPorts(SoftInfo_t **	SoftInfoTree,
				   char ** 		SearchExp)
{
    static char		Path[MAXPATHLEN];
    static char		File[MAXPATHLEN];
    char **		List;
    char **		Name;
    char *		cp = NULL;
    char *		Pkg;
    char *		Version;

    McDebug("BuildSoftInfoPorts: Scanning <%s>", PKG_DIR);
    List = DirGetNames(PKG_DIR, FALSE, &cp);
    if (!List) {
	McDebug("BuildSoftInfoPorts: No packages found in <%s>: %s",
		PKG_DIR, PRTS(cp));
	return (Opaque_t) -1;
    }

    /*
     * Each Name should be of format "pkg-version"
     */
    for (Name = List; Name && *Name; ++Name) {
	McDebug("BuildSoftInfoPorts: Checking <%s>", *Name);
	/* Make a copy of *Name we can modify */
	snprintf(File, sizeof(File), "%s", *Name);
	Pkg = File;
	Version = NULL;
	if (cp = strchr(Pkg, '-')) {
	    *cp = CNULL;
	    Version = cp + 1;
	}

	if (!Pkg || !Version) {
	    McDebug("BuildSoftInfoPorts:\t Skipping <%s>", *Name);
	    continue;
	}

	snprintf(Path, sizeof(Path), "%s/%s", PKG_DIR, *Name);

	SetPorts(Path, Pkg, Version, SoftInfoTree);
    }

    DestroyArgv(&List, -1);

    return (Opaque_t) 0;
}
