/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/5.0/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.7 $";
#endif

/*
 * MacOS functions for building softinfo using Bundle info
 */

#include "defs.h"
#include "iok.h"
#include <dirent.h>

/*
 * Directories to search for Bundle Applications in.
 * Each directory is recursively searched.
 */
static char *BundleDirs[] = {
    "/Applications",
    "/Developer/Applications", NULL
};

/*
 * Relative paths for valid bundle directories
 */
static char *BundleTestPaths[] = {
    "Contents/Info.plist", "Contents/Info-macos.plist",
    "Resources/version.plist",
    NULL
};

/*
 * Set master for SoftInfo
 * --RECURSE--
 */
static int BundleMaster(SoftInfo_t **	SoftInfoTree,
			SoftInfo_t *	SoftInfo)
{
    SoftInfo_t *		Master;
    SoftInfoFind_t		Find;
    char *			MasterName;
    char *			cp;

    if (!SoftInfo || !SoftInfo->Name || !SoftInfoTree)
	return -1;

    if (SoftInfo->Master)
	return 0;

    MasterName = strdup(SoftInfo->Name);
    cp = strrchr(MasterName, '.');
    if (!cp) {
	SImsg(SIM_DBG, "BundleMaster: Package <%s> has no '.' in it.",
	      SoftInfo->Name);
	(void) free(MasterName);
	return -1;
    }
    *cp = CNULL;

    SImsg(SIM_DBG, "BundleMaster: Look for master <%s> for <%s>",
	  MasterName, SoftInfo->Name);

    /*
     * XXX If more than one version of MasterName exists, this code finds
     * only the first version.
     */
    (void) memset(&Find, 0, sizeof(Find));
    Find.Name = MasterName;
    Find.Tree = *SoftInfoTree;
    Find.Expr = DFE_OR;
    Master = SoftInfoFind(&Find);

    if (!Master) {
	/*
	 * Create Master as a PRODUCT bundle
	 */
	SImsg(SIM_DBG, "BundleMaster: Creating master <%s> for <%s>",
	      MasterName, SoftInfo->Name);
	Master = SoftInfoCreate(NULL);
	Master->Name = MasterName;
	Master->EntryType = MC_SET_PROD_S;
	Master->EntryTypeNum = MC_SET_PROD;
	/* Recurse - add the master */
	BundleMaster(SoftInfoTree, Master);
	/* Add to tree */
	SoftInfoAdd(Master, SoftInfoTree, NULL);
    }

    SoftInfo->Master = Master;

    return 0;
}

/*
 * Get dictionary info for myBundle
 */
static int BundleGetInfo(CFBundleRef	myBundle,
			 IOKparam_t *	Params)
{
    SoftInfo_t **		SoftInfoTree;
    char *			Dir;
    SoftInfo_t *		SoftInfo = NULL;
    CFDictionaryRef 		bundleInfoDict;
    int				Status = 0;
    CFStringRef			sref;
    char *			cp;
    int				l;

    if (!Params)
	return -1;

    SoftInfoTree = Params->SoftInfoTree;
    Dir = Params->Dir;

    // Get an instance of the info plist.
    bundleInfoDict = CFBundleGetInfoDictionary( myBundle );
    if (!bundleInfoDict) {
	SImsg(SIM_DBG, "BundleGetInfo: 0x%x Get info dictionary failed",
	      myBundle);
	return -1;
    }

    SoftInfo = SoftInfoCreate(NULL);

    if (sref = CFBundleGetIdentifier(myBundle))
	if (cp = IOKstring(sref))
	    SoftInfo->Name = strdup(cp);

    if (!SoftInfo->Name) {
	SImsg(SIM_DBG, "BundleGetInfo: 0x%x Cannot find bundle name",
	      myBundle);
	return -1;
    }

    if (Dir && !SoftInfo->BaseDir)
	SoftInfo->BaseDir = strdup(Dir);

    SImsg(SIM_DBG, "BundleGetInfo: Found <%s>", PRTS(SoftInfo->Name));

    // Add info from bundleInfoDict to SoftInfo
    Params->SoftInfo = SoftInfo;
    CFDictionaryApplyFunction(bundleInfoDict, IOKSetProp, Params);

    // Cleanup SoftInfo
    if (SoftInfo->Desc && SoftInfo->Version) {
	// If Desc looks like "1.2, Mail App..." then drop the "1.2, " part
	l = strlen(SoftInfo->Version);
	if (eqn(SoftInfo->Desc, SoftInfo->Version, l)) {
	    SoftInfo->Desc += l;
	    if (eqn(SoftInfo->Desc, ", ", 2))
		SoftInfo->Desc += 2;
	}
	if (!*SoftInfo->Desc)
	    SoftInfo->Desc = NULL;
    }
    // Remove Copyright (if any) from Desc
    if (REmatch(SoftInfo->Desc, "Copyright", NULL)) {
	// The easy case is when the Desc starts with "Copyright"
	if (EQN(SoftInfo->Desc, "Copyright", 9)) {
	    SoftInfo->Copyright = SoftInfo->Desc;
	    SoftInfo->Desc = NULL;
	} else {
	    // Look for "Copyright" and zap it all
	    for (cp = SoftInfo->Desc; cp && *cp; ++cp)
		if (*cp == 'c' || *cp == 'C')
		    if (EQN(cp, "Copyright", 9)) {	// Found it
			SoftInfo->Copyright = cp;
			*(cp-1) = CNULL;		// Backup
			if (*(cp-2) == ',')
			    *(cp-2) = CNULL;		// Backup
		    }
	}
    }
    // Set Vendor from Name if we have 'com.vendor.something'
    if (!SoftInfo->VendorName && (cp = strchr(SoftInfo->Name, '.'))) {
	SoftInfo->VendorName = strdup(cp + 1);
	if (cp = strchr(SoftInfo->VendorName, '.'))
	    *cp = CNULL;
	Capitalize(SoftInfo->VendorName);
    }

    // Fix BaseDir if it was set to something not a dir
    if (SoftInfo->BaseDir && !IsDir(SoftInfo->BaseDir, 0))
	if (cp = DirName(SoftInfo->BaseDir)) {
	    (void) free(SoftInfo->BaseDir);
	    SoftInfo->BaseDir = strdup(cp);
	}

    // XXX Calculate disk usage (this can be a bit slow).
    // On a default MacOS 10.2.3 system this added about 4 seconds real time.
    if (SoftInfo->BaseDir)
	SoftInfo->DiskUsage = DiskUsage(SoftInfo->BaseDir, 0);

    // Set Master info
    BundleMaster(SoftInfoTree, SoftInfo);

    // Add to tree
    SoftInfoAdd(SoftInfo, SoftInfoTree, NULL);

    return Status;
}

/*
 * Get Bundle info from Dir and add it to SoftInfoTree
 */
static int BundleGet(SoftInfo_t **	SoftInfoTree,
		     char *		Dir)
{
    CFURLRef    		bundleURL;
    CFBundleRef 		myBundle;
    int				Status = 0;
    IOKparam_t			Params;

    SImsg(SIM_DBG, "BundleGet: Reading <%s>", Dir);

    // Make a CFURLRef from the CFString representation of the 
    // bundle's path. See the Core Foundation URL Services chapter 
    // for details.

    bundleURL = CFURLCreateWithFileSystemPath( 
				      kCFAllocatorDefault, 
				      CFStringCreateWithCString(
						NULL,
						Dir,
						CFStringGetSystemEncoding()),
				      kCFURLPOSIXPathStyle,
				      true );
    if (!bundleURL) {
	SImsg(SIM_DBG, "BundleGet:\tFailed to create bundle URL");
	return -1;
    }

    // Make a bundle instance using the URLRef.
    myBundle = CFBundleCreate( kCFAllocatorDefault, bundleURL );
    if (!myBundle) {
	SImsg(SIM_DBG, "BundleGet:\tFailed to create Bundle reference");
	return -1;
    }

    // Get Dictionary Info for this bundle
    memset(&Params, 0, sizeof(Params));
    Params.SoftInfoTree = SoftInfoTree;
    Params.Dir = Dir;
    Status = BundleGetInfo(myBundle, &Params);

    // Any CF objects returned from functions with "create" or 
    // "copy" in their names must be released by us!
    CFRelease( bundleURL );
    CFRelease( myBundle );

    return Status;
}

/*
 * Scan a Bundle Directory and call BundleGet() on each BundleDir found.
 * --RECURSE--
 */
static int BundleDirScan(SoftInfo_t **	SoftInfoTree,
			 char *		BundleDir)
{
    DIR *			dp;
    struct dirent *		Dir;
    char			Name[MAXPATHLEN];
    char			Path[MAXPATHLEN];
    char			InfoFile[MAXPATHLEN];
    int				Found = 0;
    char **			cpp;

    SImsg(SIM_DBG, "BundleDirScan: Scanning <%s> for bundle apps", BundleDir);

    dp = opendir(BundleDir);
    if (!dp) {
	SImsg(SIM_DBG, "BundleDirScan: %s: Open dir failed: %s",
	      BundleDir, SYSERR);
	return -1;
    }

    while (Dir = readdir(dp)) {
	snprintf(Name, sizeof(Name), "%.*s", Dir->d_namlen, Dir->d_name);
	if (eq(Name, ".") || eq(Name, ".."))
	    continue;

	snprintf(Path, sizeof(Path), "%s/%s", BundleDir, Name);

	if (!IsDir(Path, FF_SILENCE))
	    continue;

	/*
	 * These FileExists() tests are a performance enhancement.
	 * It makes us about 35% faster doing this.
	 */
	for (cpp = BundleTestPaths; cpp && *cpp; ++cpp) {
	    snprintf(InfoFile, sizeof(InfoFile), "%s/%s", Path, *cpp);
	    if (FileExists(InfoFile, FF_SILENCE)) {
		if (BundleGet(SoftInfoTree, Path) == 0)
		    ++Found;
		break;
	    }
	}

	/* Descend */
	BundleDirScan(SoftInfoTree, Path);
    }

    closedir(dp);

    return (Found > 0) ? 0 : -1;
}

/*
 * Build Software Information from Bundles
 */
extern int BuildSoftInfoBundle(SoftInfo_t **	SoftInfoTree,
			       char **		SearchExp)
{
    char **			cpp;
    int				Status = -1;
    CFArrayRef			Bundles;
    CFRange			Range;
    IOKparam_t			Params;

    SImsg(SIM_DBG, "BuildSoftInfoBundle: Scanning Bundles for software info");

    if (Bundles = CFBundleGetAllBundles()) {
	CFRange    		Range = {0, CFArrayGetCount(Bundles) };

	SImsg(SIM_DBG, "BuildSoftInfoBundle: #Bundles=%d", Range.length);

	memset(&Params, 0, sizeof(Params));
	Params.SoftInfoTree = SoftInfoTree;
	CFArrayApplyFunction(Bundles,
			     Range,
			     (CFArrayApplierFunction) BundleGetInfo,
			     (void *) &Params);

	/* Test to see if we did something */
	if (SoftInfoTree && *SoftInfoTree)
	    Status = 0;
    } else {
	SImsg(SIM_DBG, "BuildSoftInfoBundle: No Bundles were found");
    }

    for (cpp = BundleDirs; cpp && *cpp; ++cpp)
	if (BundleDirScan(SoftInfoTree, *cpp) == 0)
	    Status = 0;

    return Status;
}
