/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/5.0/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.3 $";
#endif

/*
 * MacOS X SCSI
 */

#include "defs.h"
#include "myscsi.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/errno.h> 
#include <sys/fcntl.h>
#include <sys/stat.h>   
#include <mach/mach.h>
#include <Carbon/Carbon.h>
#include <CoreFoundation/CoreFoundation.h>
#include <IOKit/IOKitLib.h>
#include <IOKit/IOCFPlugIn.h>
#include <IOKit/cdb/IOSCSILib.h>
#include <unistd.h>
#include <sysexits.h>
#include <mach/mach_error.h>

/*
 * Create an IOK SCSI Interface to allow the caller to speak SCSI to DevInfo
 */
static int SCSICreateInterface(DevInfo_t *DevInfo,
			       const io_object_t scsiDevice,
			       IOSCSIDeviceInterface ***scsiDeviceInterface
			       )
{
    io_name_t 			className;
    IOCFPlugInInterface **	plugInInterface = NULL;
    HRESULT 			plugInResult = S_OK;
    IOReturn 			ioReturnValue = kIOReturnSuccess;
    SInt32 			score = 0;

    if (!DevInfo)
	return -1;

    ioReturnValue = IOObjectGetClass(scsiDevice, className);
    SImsg(SIM_DBG, "SCSICreateInterface: <%s> Class=<%s>",
	  DevInfo->Name, className);

    // Create an intermediate interface we'll use to create the device interface.
    ioReturnValue = IOCreatePlugInInterfaceForService(scsiDevice,
						      kIOSCSIUserClientTypeID,
						      kIOCFPlugInInterfaceID,
						      &plugInInterface,
						      &score);
    if (ioReturnValue != kIOReturnSuccess) {
	SImsg(SIM_DBG,
	      "SCSICreateInterface: <%s> Couldn't create a plug-in interface: %s",
	      DevInfo->Name, IOKerr(ioReturnValue));
	return -1;
    }

    // Call function of intermediate plug-in to create device interface.
    plugInResult = (*plugInInterface)->QueryInterface(plugInInterface, 
			      CFUUIDGetUUIDBytes(kIOSCSIDeviceInterfaceID),
                              (LPVOID) scsiDeviceInterface);
    if (plugInResult != S_OK) {
	SImsg(SIM_DBG,
	      "SCSICreateInterface: <%s> Couldn't create SCSI device interface",
	      DevInfo->Name);
	return -1;
    }

    // We're now finished with the instance of IOCFPlugInInterface.
    (*plugInInterface)->Release(plugInInterface);

    return 0;
}

/*
 * Create a CDB interface to the IOK
 */
static IOCDBCommandInterface **MyAllocCDBCommand(DevInfo_t *DevInfo,
				 IOSCSIDeviceInterface **scsiDeviceInterface)
{
    HRESULT plugInResult = S_OK;
    IOCDBCommandInterface **cdbCmdInterface = NULL;

    SImsg(SIM_DBG, "MyAllocCDBCommand: <%s> Creating CDB Command",
	  DevInfo->Name);

    plugInResult = (*scsiDeviceInterface)->QueryInterface(
                        scsiDeviceInterface,
                        CFUUIDGetUUIDBytes(kIOCDBCommandInterfaceID), 
                        (LPVOID) &cdbCmdInterface);

    if (plugInResult != S_OK) {
	SImsg(SIM_DBG,
	      "MyAllocCDBCommand: <%s> Couldn't create a CDB Command",
	      DevInfo->Name);
	return (IOCDBCommandInterface **) NULL;
    }

    return cdbCmdInterface;
}

/*
 * Darwin/MacOS X function to issue a SCSI command.
 */
extern int ScsiCmd(ScsiCmd_t *Cmd,
		   ScsiQuery_t *Query)
{
    IOSCSIDeviceInterface **	scsiDeviceInterface = NULL;
    IOCDBCommandInterface **	cdbCommandInterface = NULL;
    IOReturn 			ioReturnValue = kIOReturnSuccess;
    UInt8 			inquiryData[kDefaultInquirySize];
    IOVirtualRange 		range[1];
    UInt32 			seqNumber;
    SCSIResults 		results;

    if (!Cmd || !Query) {
	SImsg(SIM_DBG, "ScsiCmd: Missing params.");
	return -1;
    }

    // Retrieve interface created by IOKprobeSCSI()
    scsiDeviceInterface = (IOSCSIDeviceInterface **) Query->DevPort;
    if (!scsiDeviceInterface) {
	SImsg(SIM_DBG, "ScsiCmd: Device=<%s> no SCSI Interface found",
	      Query->DevInfo->Name);
	return -1;
    }

    SImsg(SIM_DBG, "ScsiCmd: Device=<%s> Interface=0x%x",
	  Query->DevInfo->Name, scsiDeviceInterface);

    //
    // Open the device, attempt to get live information for it, then close it.
    // XXX Research seems to indicate this will fail for most device types
    // as the OS wants to prevent apps from direct SCSI access.  This seems
    // to be the case as of MacOS 10.2.2 at least.
    //
    ioReturnValue = (*scsiDeviceInterface)->open(scsiDeviceInterface);
    if ( ioReturnValue != kIOReturnSuccess ) {
	SImsg(SIM_DBG, "ScsiCmd: <%s> failed to open interface: %s",
	      Query->DevInfo->Name, IOKerr(ioReturnValue));
	return -1;
    }

    // Create CDB command interface and execute a command to query the device.
    cdbCommandInterface = MyAllocCDBCommand(Query->DevInfo,
					    scsiDeviceInterface);
    if (!cdbCommandInterface) {
	SImsg(SIM_DBG, "ScsiCmd: <%s> failed to create cdb interface",
	      Query->DevInfo->Name);
	return -1;
    }

    // Prep for CDB query
    range[0].address = (IOVirtualAddress) inquiryData;
    range[0].length  = sizeof(inquiryData);

    // Execute the CDB query
    ioReturnValue = (*cdbCommandInterface)->setAndExecuteCommand(
                            cdbCommandInterface,
                            Cmd->Cdb,
			    Cmd->CdbLen,
                            range,
                            sizeof(range)/sizeof(range[0]),
                            0, 				/* isWrite */
                            MySCSI_CMD_TIMEOUT * 1000, 	/* timeoutMS */
                            0, 				/* target */
                            0, 				/* callback */
                            0, 				/* refcon */
                            &seqNumber);
    if (ioReturnValue != kIOReturnSuccess) {
        if (ioReturnValue != kIOReturnUnderrun) {
            SImsg(SIM_DBG, "ScsiCmd: <%s> Couldn't execute CDB command: %s",
		  Query->DevInfo->Name, IOKerr(ioReturnValue));
            return -1;
        } else {   // alert developer but don?t stop running
            SImsg(SIM_DBG, "ScsiCmd: <%s> Command underrun: %s",
		  Query->DevInfo->Name, IOKerr(ioReturnValue));
	}
    }
    // Fetch the results
    ioReturnValue = (*cdbCommandInterface)->getResults(cdbCommandInterface, 
						       &results);
    if (ioReturnValue != kIOReturnSuccess) {
        if (ioReturnValue != kIOReturnUnderrun) {
            SImsg(SIM_DBG,
		  "ScsiCmd: <%s> Couldn't get results of a command: %s",
		  Query->DevInfo->Name, IOKerr(ioReturnValue));
            return -1;
        } else {   // alert developer but don?t stop running
            SImsg(SIM_DBG, "ScsiCmd: <%s> getResults Underrun",
		  Query->DevInfo->Name);
	}
    }

    // Save results
    Cmd->Data = (void *) inquiryData;

    // Free the cdb interface
    (*cdbCommandInterface)->Release(cdbCommandInterface);
    // Close the device we opened.
    (*scsiDeviceInterface)->close(scsiDeviceInterface);

    return 0;
}

/*
 * Used OS cached SCSI info
 */
static int ScsiQueryCached(DevInfo_t *DevInfo,
			   IOSCSIDeviceInterface **	scsiDeviceInterface)
{
    UInt8 			inquiryData[255];
    UInt32 			inquiryDataSize = sizeof(inquiryData);
    IOReturn 			ioReturnValue = kIOReturnSuccess;
    static ScsiQuery_t	        Query;
    int				Status = 0;

    if (!DevInfo || !scsiDeviceInterface)
	return -1;

    SImsg(SIM_DBG,
	  "ScsiQueryCached: Device <%s> Looking for OS cached SCSI data",
	  PRTS(DevInfo->Name));

    bzero(inquiryData, sizeof(inquiryData));    // Zero data block.

    SImsg(SIM_DBG, "ScsiQueryCached:\tFetching cached INQUIRY data");

    // Call a function of the IOSCSIDeviceInterface that returns cached
    // information about the device.
    ioReturnValue = (*scsiDeviceInterface)->getInquiryData(
                            scsiDeviceInterface, (SCSIInquiry *) inquiryData, 
                            sizeof(inquiryData), &inquiryDataSize);

    if (ioReturnValue != kIOReturnSuccess) {
	SImsg(SIM_DBG, "ScsiQueryCached:\tCouldn't get inquiry data: %s",
	      IOKerr(ioReturnValue));
	return -1;
    }

    memset(&Query, 0, sizeof(Query));
    Query.DevInfo = DevInfo;
    Query.Data = (void *) inquiryData;
    Query.DataLen = (size_t) inquiryDataSize;

    if (ScsiInquiryDecode(&Query) < 0)
	--Status;

    return Status;
}

/*
 * Probe a SCSI device
 */
extern int IOKprobeSCSI(DevInfo_t *DevInfo,
			io_registry_entry_t Service)
{
    io_object_t 		scsiDevice = NULL;
    IOSCSIDeviceInterface **	scsiDeviceInterface = NULL;
    ScsiQuery_t			Query;
    int				Status = 0;

    SImsg(SIM_DBG, "IOKprobeSCSI: Device <%s>",
	  PRTS(DevInfo->Name));

    /* Create an Iterator from the Service */
    scsiDevice = (io_object_t) Service;
#if 0
    /* XXX This appears unnecessary */
    if (IORegistryEntryCreateIterator(Service,
				      /*kIODeviceTreePlane,*/
				      kIOServicePlane,
				      0,
				      &scsiDevice) != kIOReturnSuccess) {
	SImsg(SIM_DBG,
	      "IOKprobeSCSI:\tCreate Iterator from Registry failed.",
	      PRTS(DevInfo->Name));
	return -1;
    }
#endif

    /* Create an IO Kit Interface (scsiDeviceInterface) */
    if (SCSICreateInterface(DevInfo,
			    scsiDevice,
			    &scsiDeviceInterface) < 0)
	return -1;

    SImsg(SIM_DBG, "IOKprobeSCSI:\tInterface=0x%x",
	  PRTS(DevInfo->Name), scsiDeviceInterface);

    /*
     * Perform full suite of SCSI queries on live device.
     * XXX Research seems to indicate that the OS (10.2.2) refuses
     * to allow an application (such as this one) from opening
     * a device directly.  This prevents us from performing a full
     * suite of SCSI queries.  Hopefully this will change in a future
     * release.
     */
    memset(&Query, CNULL, sizeof(Query));
    Query.DevPort = (Opaque_t) scsiDeviceInterface;
    Query.DevInfo = DevInfo;
    Status = ScsiQueryDirect(&Query);
    SImsg(SIM_DBG, "IOKprobeSCSI:\tSCSI Query Result=%d",
	  PRTS(DevInfo->Name), Status);

    if (Status < 0) {
	/* Full SCSI Query suite failed so pull out basic cached data */
	SImsg(SIM_DBG, "IOKprobeSCSI:\tFull SCSI Query failed");
	Status = ScsiQueryCached(DevInfo, scsiDeviceInterface);
    }

    return Status;
}
