/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/5.0/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.15 $";
#endif

/*
 * SysInfo Server
 */

#include "defs.h"
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>

#define MAX_NULL_INPUTS		10	/* Max times client sends nothing */

/*
 * Alarm handler
 */
extern void SigAlarm(int Sig)
{
    /* Turn off alarm */
    signal(SIGALRM, SIG_IGN);
    /* Report */
    Debug("SigAlarm: Timeout has occured.");
    SImsg(SIM_INFO, "session timed out\n");
    /* Say goodbye */
    exit(1);
}

/*
 * SIGPIPE handler - called when client is disconnected
 */
extern void SigPipe(int Sig)
{
    /* Turn off so we don't loop */
    signal(SIGPIPE, SIG_IGN);
    /* Report */
    Debug("SigPipe: SIGPIPE received");
    SImsg(SIM_INFO, "session disconnected\n");
    /* Say goodbye */
    exit(1);
}

/*
 * Service Client requests
 */
extern int ClientService(int Timeout)
{
    SSPsession_t *		s;
    char *			Line;
    char *			CmdName;
    int				Argc;
    char **			Argv;
    CmdDispatch_t *		Cmd;
    int				BadCnt = 0;
    int				EmptyCnt = 0;

    /*
     * Setup timeout handler 
     */
    if (Timeout) {
	(void) signal(SIGALRM, SigAlarm);
	(void) alarm(Timeout);
	Debug("ClientService: Timeout set to %d seconds", Timeout);
    }

    /*
     * Handle client disconnects
     */
    signal(SIGPIPE, SigPipe);

    /*
     * Setup the session and display welcome banner
     */
    s = SSPstart();
    if (!s)
	return -1;

    Debug("ClientService: Entering client command service loop");

    /*
     * Main command service loop
     */
    for ( ; ; ) {
	if (Timeout) {
	    /* Reset alarm */
	    (void) alarm(Timeout);
	}

	/* Read command input from client */
	Line = SSPreadLine(s);

	/* Check to see if SSPreadLine shutdown the session */
	if (!SSPsessionValid(s)) {
	    Debug("ClientService: Session is no longer valid.");
	    return 0;
	}

	if (!Line) {
	    /*
	     * Check to insure we don't go into a loop or hammer the 
	     * server system due to a client sending lots of null inputs.
	     */
	    if (EmptyCnt++ > MAX_NULL_INPUTS) {
		SImsg(SIM_WARN,
		      "Max # (%d) of null/empty inputs from client reached.",
		      MAX_NULL_INPUTS);
		SSPclose(s);
		exit(1);
	    }
	    continue;
	}

	/* Debug info */
	SImsg(SIM_DBG, "< %s", Line);

	/* Log the command */
	SImsg(SIM_INFO, "%s command is %s\n", SSPclientStr(s), PRTS(Line));
	      
	Argc = StrToArgv(Line, " ", &Argv, NULL, 0);

	if (Argc <= 0) {
	    SSPsend(s, SSP_RC_ERR, SSP_RCC_SINGLE, "No command was received");
	    /*
	     * Slow down responses at a given threshold to downplay
	     * Denial-Of-Service attacks or runaway clients.
	     */
	    if (BadCnt++ > 10) {
		SImsg(SIM_WARN, "Bad command count now %d; sleeping", BadCnt);
		sleep(5);
	    }
	    continue;
	}
	CmdName = Argv[0];

	/* Lookup the command */
	Cmd = SSPcmdGet(s, CmdName);
	if (!Cmd) {
	    SSPsend(s, SSP_RC_ERR, SSP_RCC_SINGLE,
		    "Command \"%s\" was not recognized",
		    Argv[0]);
	    DestroyArgv(&Argv, Argc);
	    continue;
	}

	/* Check to see if authentication is first required */
	if (Cmd->AuthReq && s->AuthState != AUTH_STATE_GOOD) {
	    SSPsend(s, SSP_RC_AUTHREQ, SSP_RCC_SINGLE,
		    "You must authenticate with %s first",
		    SSP_CMD_AUTH);
	    DestroyArgv(&Argv, Argc);
	    continue;
	}

	/*
	 * Execute the command.
	 * Each Func() is responsible for responding to client.
	 */
	(void) (*Cmd->Func)(s, Argv, Argc, Cmd);

	DestroyArgv(&Argv, Argc);
    }

    return 0;
}

/*
 * Initialize a client connection on SockFdClient
 */
extern int ClientInit(int		SockFdClient,
		      int		SockFd,
		      int		Timeout)
{
    int				f;
    pid_t			Pid;

    /*
     * Make the socket blocking
     */
    f = fcntl(SockFdClient, F_GETFL, 0);
    if (fcntl(SockFdClient, F_SETFL, f & ~O_NONBLOCK) < 0) {
	SImsg(SIM_CERR, "Set client socket connect to be blocking failed: %s",
	      SYSERR);
	return -1;
    }

    if (ForkFlag) {
	Pid = fork();
	if (Pid < 0) {
	    SImsg(SIM_CERR, "ClientInit(): fork failed: %s", SYSERR);
	    return -1;
	}
    }

    if (!ForkFlag || Pid == 0) {
	/* I'm the child */
	Debug("ClientInit: Preparing to handle connection");
	    
	/* Children shouldn't play with signals */
	signal(SIGCHLD, SIG_DFL);
	signal(SIGTERM, SIG_DFL);
	signal(SIGHUP, SIG_DFL);
	
	/* Don't need our parent */
	close(SockFd);
	SockFd = -1;
	
	/*
	 * Setup descriptors so that we talk to stdin/stdout
	 */
	dup2(SockFdClient, 0);		/* stdin */
	dup2(SockFdClient, 1);		/* stdout */
	close(SockFdClient);
	SockFdClient = -1;

	/*
	 * Process Service Commands
	 */
	ClientService(Timeout);

	exit(0);
    } else {
	/* I'm the parent */
	Debug("ClientInit: Forked child %d", Pid);
	close(SockFdClient);
    }

    return 0;
}
