/*
 * Copyright (c) 1992-2003 MagniComp
 * This software may only be used in accordance with the license which is
 * available as http://www.magnicomp.com/sysinfo/5.0/sysinfo-eu-license.shtml
 */

#ifndef lint
static char *RCSid = "$Revision: 1.15 $";
#endif

/*
 * SysInfo Engine (SIE) functions
 */

#include "defs.h"

/*
 * SIE option entry
 */
typedef struct {
    char *			Name;		/* Option name */
    char *			SieOption;	/* Corresponding CLI option */
    char *			Arg;		/* Caller supplied arg */
    int				Enabled;	/* Is this option enabled? */
} SieOption_t;

/*
 * List of acceptable SIE options
 */
static SieOption_t SieOptions[] = {
    { "licports",	"--licports" },
    { "msglevel",	"--msglevel" },
    { "noserial", 	"--noserial" },
    { "swfiles", 	"+swfiles" },
    { 0 },
};

/*
 * Expand Key and add to Buffer
 */
static int ExpandKey(char *	Key,
		     char **	BufferPtr,
		     int	BufLen)
{
    char *			Val = NULL;

    if (!Key || !BufferPtr)
	return -1;

    if (EQ(Key, "CFDIR"))
	Val = CFDIR;
    else if (EQ(Key, "RUNDIR"))
	/* Directory we were run from */
	Val = DirName(ProgPath);

    if (Val) {
	snprintf(*BufferPtr, BufLen, "%s", Val);
	*(BufferPtr) += strlen(Val);
	return 0;
    }

    return -1;
}

/*
 * Expand Path by looking for ${...} variables and return expanded result.
 */
static char *Expand(char *	Path)
{
    static char			NewPath[MAXPATHLEN];
    char *			Key;
    char *			cp;
    char *			np;

    memset(NewPath, CNULL, sizeof(NewPath));

    for (cp = Path, np = NewPath; cp && *cp; ) {
	if (*cp == '$' && *(cp+1) == '{') {
	    Key = cp + 2;
	    cp = strchr(Key, '}');
	    if (!cp) {
		Debug("Expand: No matching '}' found: %s", Path);
		return (char *) NULL;
	    }
	    *cp = CNULL;
	    ++cp;
	    if (ExpandKey(Key, &np, sizeof(NewPath) - strlen(np)) < 0) {
		SImsg(SIM_GERR, 
		      "Expand: Key ${%s} is not valid in %s", Key, Path);
		return (char *) NULL;
	    }
	} else {
	    *np++ = *cp++;
	}
    }

    return NewPath;
}

/*
 * Find the path to the "sysinfo" program
 */
extern char *SIEsysinfoPath(char	*UserPath)
{
    static char *		Sie;
    char *			SearchPath;
    char *			Path;
    int				Argc;
    char **			Argv;
    int				i;

    if (Sie)
	return Sie;

    if (UserPath) {
	if (!FileExists(UserPath, 0)) {
	    SImsg(SIM_CERR, "SysInfo command \"%s\" does not exist.",
		  UserPath);
	    return (char *) NULL;
	}

	Sie = strdup(UserPath);
	Debug("SIEsysinfoPath: Will use sysinfo(1) command <%s>", Sie);
	return Sie;
    }

    SearchPath = CFgetValue(CF_SYSINFOPATH);
    if (!SearchPath)
	SearchPath = SYSINFO_PATH;

    Argc = StrToArgv(SearchPath, ":", &Argv, NULL, 0);
    for (i = 0; i < Argc; ++i) {
	Path = Expand(Argv[i]);
	if (!Path)
	    continue;
	Debug("SIEsysinfoPath: Checking for <%s>", Path);
	if (FileExists(Path, 0)) {
	    Sie = strdup(Path);
	    Debug("SIEsysinfoPath: Found sysinfo(1) command <%s>", Sie);
	    return Sie;
	}
    }

    SImsg(SIM_CERR, "Could not find the sysinfo(1) program.");

    return (char *) NULL;
}

/*
 * Set OptName in the Sie Options table
 */
extern int SIEoptionSet(char *OptName)
{
    SieOption_t *		op;
    char *			cp;
    char *			Arg = NULL;

    if (!OptName)
	return -1;

    /*
     * Look for "optname=arg"
     */
    if (cp = strchr(OptName, '=')) {
	*cp = CNULL;
	Arg = cp + 1;
    }

    Debug("SIEoptionSet: OptName=<%s> Arg=<%s>", PRTS(OptName), PRTS(Arg));

    for (op = SieOptions; op && op->Name; ++op) {
	if (!EQ(OptName, op->Name))
	    continue;
	/* Found it */
	op->Enabled = TRUE;
	if (op->Arg)			/* Check for previously set Arg */
	    (void) free(op->Arg);	/* Cleanup from previous call */
	if (Arg)
	    op->Arg = strdup(Arg);

	return 0;			/* Success */
    }

    Debug("SIEoptionSet: Option <%s> is not recognized", OptName);

    return -1;				/* Failure */
}

/*
 * Return a string of base SIE options to use.
 */
extern char *SIEoptionStr()
{
    static char			Options[256];
    char			Opt[128];
    SieOption_t *		op;

    snprintf(Options, sizeof(Options), "%s ", SYSINFO_BASE_ARGS);

    for (op = SieOptions; op && op->Name; ++op) {
	if (op->Enabled) {
	    if (op != SieOptions)
		StrAppend(Options, sizeof(Options), " ");
	    StrAppend(Options, sizeof(Options), op->SieOption);
	    if (op->Arg) {
		StrAppend(Options, sizeof(Options), " ");
		StrAppend(Options, sizeof(Options), op->Arg);
	    }
	}
    }

    return Options;
}

/*
 * Return list of supported class names.
 */
extern char **SIEclassList()
{
    char *			SysInfo;
    static char			Line[512];
    static char			Cmd[MAXPATHLEN];
    static char	*		List[MAX_CLASSLIST];
    char *			cp;
    static int			ListCnt;
    FILE *			PH;

    if (ListCnt)
	return List;

    SysInfo = SIEsysinfoPath(NULL);
    if (!SysInfo)
	return (char **) NULL;

    snprintf(Cmd, sizeof(Cmd), "%s %s --list class --msglevel terse",
	     SysInfo, SIEoptionStr());
    Debug("SIEclassList: Run: %s", Cmd);

    PH = popen(Cmd, "r");
    if (!PH) {
	SImsg(SIM_GERR, "Open pipe to command \"%s\" failed: %s",
	      Cmd, SYSERR);
	return (char **) NULL;
    }

    /*
     * Lines look like:
     *	General              General Information
     *	Device,Hardware      Hardware Information
     */
    while (fgets(Line, sizeof(Line), PH)) {
	/* Find the end of the first word which is the class name */
	for (cp = Line; cp && *cp && *cp != ',' && !isspace(*cp); ++cp);
	/* Zap it */
	*cp = CNULL;
	if (strlen(Line) <= 0)
	    continue;

	if (ListCnt >= MAX_CLASSLIST) {
	    SImsg(SIM_GERR, "SIEclassList: MAX_CLASSLIST (%d) limit reached.",
		  MAX_CLASSLIST);
	    return List;
	}
	if (cp = strlower(Line))		/* Make lower case */
	    List[ListCnt++] = strdup(cp);
    }

    pclose(PH);

    return (ListCnt) ? List : (char **) NULL;
}

/*
 * Return version of SysInfo command
 */
extern char *SIEversion()
{
    char *			SysInfo;
    static char			Line[512];
    static char			Cmd[MAXPATHLEN];
    static char	*		Version;
    char *			cp;
    FILE *			PH;

    if (Version)
	return Version;

    SysInfo = SIEsysinfoPath(NULL);
    if (!SysInfo)
	return (char *) NULL;

    snprintf(Cmd, sizeof(Cmd), "%s %s --msglevel terse --version",
	     SysInfo, SIEoptionStr());
    Debug("SIEversion: Run: %s", Cmd);

    PH = popen(Cmd, "r");
    if (!PH) {
	SImsg(SIM_GERR, "Open pipe to command \"%s\" failed: %s",
	      Cmd, SYSERR);
	return (char *) NULL;
    }

    if (fgets(Line, sizeof(Line), PH)) {
	/* Zap newline */
	if (cp = strchr(Line, '\n'))
	    *cp = CNULL;
	/* Change all whitespace to '-' */
	for (cp = Line; cp && *cp; ++cp)
	    if (isspace(*cp))
		*cp = '-';
	Version = Line;
	Debug("SIEversion: SysInfo version is <%s>", Version);
    } else {
	SImsg(SIM_GERR, "No version info read from command: %s", Cmd);
    }

    pclose(PH);

    return Version;
}

/*
 * Run SIE with Options and store results in ListPtr
 */
static int SieGetLines(char ***	ListPtr,
		       char *	Options)
{
    char *			SysInfo;
    static char			Line[8192];
    static char			Cmd[MAXPATHLEN];
    char	**		List = NULL;
    int				ListCnt = 0;
    char *			cp;
    FILE *			PH;

    if (!ListPtr)
	return -1;

    SysInfo = SIEsysinfoPath(NULL);
    if (!SysInfo)
	return -1;

    snprintf(Cmd, sizeof(Cmd), "%s %s --format report --repsep '|' %s",
	     SysInfo, SIEoptionStr(), Options);
    Debug("SieGetLines: Run: %s", Cmd);

    PH = popen(Cmd, "r");
    if (!PH) {
	SImsg(SIM_GERR, "Open pipe to command \"%s\" failed: %s",
	      Cmd, SYSERR);
	return -1;
    }

    while (fgets(Line, sizeof(Line), PH)) {
	/* Zap newline */
	if (cp = strchr(Line, '\n'))
	    *cp = CNULL;
	Debug("SieGetLines: Read SysInfo line: %s", Line);

	if (!ListCnt) {
	    List = (char **) xcalloc(1, sizeof(char *));
	    ListCnt = 1;
	} else {
	    List = (char **) xrealloc(List, ++ListCnt * sizeof(char *));
	}
	List[ListCnt-1] = strdup(Line);
    }

    pclose(PH);

    *ListPtr = List;

    return ListCnt;
}

/*
 * Return license info
 */
extern int SIElicense(char ***ListPtr)
{
    return SieGetLines(ListPtr, "--license");
}

/*
 * Return members info
 */
extern int SIEmembers(char ***ListPtr)
{
    return SieGetLines(ListPtr, "--members");
}

/*
 * Return types info
 */
extern int SIEtypes(char ***ListPtr)
{
    return SieGetLines(ListPtr, "--list type");
}
